
# from .db import Database
from collections.abc import MutableSet

class TagTable:
	"""Object representing the 'tag' table in the :class:`.Database`."""

	def __init__(self, 
		db, 
		table: str = 'tag',
	) -> None:
		
		self._db = db
		self._table = table

	### FACTORIES

	### PROPERTIES

	@property
	def db(self):
		"""Returns the associated :class:`.database`"""
		return self._db
	
	@property
	def table(self) -> str:
		return self._table

	@property
	def unique(self) -> set:
		"""Returns a set of tag names contained in the table"""
		values = self.db.select(table=self.table, query='DISTINCT tag_name', multiple=True)
		return set(v for v, in values)

	### METHODS

	### DUNDERS

	def __getitem__(self, key):
		raise NotImplementedError



class TagSet(MutableSet):
	"""Object representing a subset of the 'tag' table in the :class:`.Database` belonging to a certain :class:`.Compound` or :class:`.Pose`."""

	def __init__(self, 
		parent, # Compound or Pose
		tags=(), 
		immutable=False,
		commit=True,
	):
		
		self._elements = []
		self._immutable = immutable
		self._parent = parent

		for tag in tags:
			self.add(tag, commit=commit)
		
	### FACTORIES

	### PROPERTIES

	@property
	def tags(self) -> list:
		"""Returns the elements in this set"""
		return self._elements

	@property
	def immutable(self):
		"""Returns whether this set is immutable"""
		return self._immutable
	
	@immutable.setter
	def immutable(self,b):
		self._immutable = b

	@property
	def parent(self):
		"""Returns this set of tags parent :class:`.Compound` or :class:`.Pose`."""
		return self._parent

	@property
	def db(self):
		"""Returns the associated :class:`.database`"""
		return self.parent.db
	

	
	### DATABASE

	def _remove_tag_from_db(self, tag):
		sql = f'DELETE FROM tag WHERE tag_name="{tag}" AND tag_{self.parent.table}={self.parent.id}'
		self.db.execute(sql)

	def _clear_tags_from_db(self, tag):
		sql = f'DELETE FROM tag WHERE tag_{self.parent.table}={self.parent.id}'
		self.db.execute(sql)

	def _add_tag_to_db(self, tag, commit=True):
		payload = { 'name':tag, self.parent.table:self.parent.id }
		self.db.insert_tag(**payload, commit=commit)
			

	### METHODS

	def pop(self):
		"""Pop the last element"""
		assert not self.immutable
		return self._elements.pop()

	def discard(self, tag):
		"""Discard an element"""
		return self.discard(tag)

	def clear(self):
		"""Clear all tags"""
		self._elements = []
		self._clear_tags_from_db(self)

	def remove(self, tag):
		"""Remove an element"""
		assert not self.immutable
		if tag in self:
			i = self._elements.index(tag)
			del self._elements[i]
			self._remove_tag_from_db(tag)
		else:
			raise ValueError(f'{tag} not in {self}')

	def add(self, tag, commit=True):
		"""Add an element"""
		assert not self.immutable
		if tag not in self._elements:
			self._elements.append(tag)
			self._add_tag_to_db(tag, commit=commit)

	def glob(self, pattern):
		import fnmatch
		return fnmatch.filter(self.tags, pattern)

	### DUNDERS

	def __contains__(self, tag):
		return tag in self.tags

	def __repr__(self):
		return str(self._elements)

	def __len__(self):
		return len(self._elements)

	def __iter__(self):
		return iter(self._elements)

	def __iter__(self):
		return iter(self.tags)

	def __add__(self, other):
		raise NotImplementedError
		# return TagTable(self.tags+other)




import pycule
from .quote import Quote
import requests

import logging
# logging.getLogger("requests").setLevel(logging.WARNING)
# logging.getLogger("mcule:core").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

ENAMINE_CATALOGUES = [
	'BB',
	'SCR',
	'MADE',
	'REAL',
]

ENAMINE_V2_CATALOGUES = {
	"building-blocks": "BB",
	'screening-compounds':'SCR' ,
	'made-building-blocks':'MADE',
	'real':'REAL',
	'bioactive-compounds':'BIO',
}

ENAMINE_V2_BASE_URL = 'https://new.enaminestore.com/api/v2'

ENAMINE_V2_LEAD_TIME = {
	"BB": {
		"IN_STOCK": 5,
		"SYNTHESIS": 7,
	},
	"SCR": {
		"IN_STOCK": 5,
		"SYNTHESIS": 20,
	},
	"MADE": {
		"SYNTHESIS": 20,
	},
	"REAL": {
		"SYNTHESIS": 20,
	},
	"BIO": {
		"IN_STOCK": 5,
	},
}

import logging
logger = logging.getLogger('HIPPO')

class Quoter:
	"""Class to scrape catalogue data"""

	def __init__(self,
		supplier: str, 
		username: str = None, 
		password: str = None,
		token: str = None,
	):
		"""Currently 'Enamine' and 'MCule' are supported. Once initialised"""

		if supplier == 'Enamine':

			self.supplier = 'Enamine'
			self.batch_size = 2000
			self.catalogues = ENAMINE_CATALOGUES
			self._get_quote = self.get_enamine_quote
			self._get_batch_quote = self.get_enamine_batch_quote

			assert username
			assert password

			self.wrapper = pycule.core.EnamineWrapper(username=username, password=password)

		elif supplier == 'MCule':

			self.supplier = 'MCule'
			self.batch_size = 1000
			self._get_quote = self.get_mcule_quote
			self._get_batch_quote = self.get_mcule_batch_quote
			
			assert token

			self.wrapper = pycule.core.MCuleWrapper(authorisation_token=token)
		
		else:
			logger.error(f'Unsupported supplier: "{supplier}"')
			raise NotImplementedError


	def get_quote(self, compound, **kwargs):
		"""Get quotes for a compound"""
		return self._get_quote(compound, **kwargs)

	def get_batch_quote(self, compounds, **kwargs):
		"""Get quotes for a compound set"""
		return self._get_batch_quote(compounds, **kwargs)

	def __call__(self, compound):
		"""Get quotes for a compound"""
		return self.get_quote(compound)

	### ENAMINE

	def get_enamine_batch_quote(self, compounds, currency="USD", catalogues=None, exact=False, forms=False, analogues=False, equivalents=False):

		import time
		from .cset import CompoundSet

		n_comps = len(compounds)
		logger.var('batch size', n_comps)
		assert n_comps <= 2000

		unmatched = compounds.copy()
		matched = CompoundSet(compounds.db)
		errors = []
							  
		# perform an exact search on the batch of smiles (for each catalog)

		payload = dict(compounds=compounds.smiles, currency=currency)

		include = []
		if forms: include.append('FORMS')
		if analogues: include.append('ANALOGS')
		if equivalents: include.append('EQUIVALENTS')
		if include: payload['include'] = include

		catalogues = catalogues or ENAMINE_V2_CATALOGUES

		t_start = time.perf_counter()

		# return payload

		for catalogue in catalogues:

			if exact:
				logger.debug(f'Exact search in Enamine:{ENAMINE_V2_CATALOGUES[catalogue]}...')
			else:
				logger.debug(f'Similarity==1.0 search in Enamine:{ENAMINE_V2_CATALOGUES[catalogue]}...')

			# send the request
			t_catalogue_start = time.perf_counter()
			if exact:
				url = ENAMINE_V2_BASE_URL + f'/catalog/search/in/{catalogue}/by/SMILEs/EXACT'
			else:
				url = ENAMINE_V2_BASE_URL + f'/catalog/search/in/{catalogue}/by/SMILEs/SIMILARITY/1.00'
			response = requests.post(url=url, json=payload)
			logger.var(f'{catalogue} request time = ', time.perf_counter() - t_catalogue_start)

			# process the request
			if (status := response.status_code) != 200:
				logger.error(f'Request error: status={status}')
				errors.append(dict(catalogue=catalogue, status=status, response=response))
				continue
			
			data = response.json()

			results = data['results']
			products = results['products']
			
			# register the quotes in the database
			this_matched = self.parse_enamine_bulk_products(compounds.db, products)

			if this_matched:
				logger.success(f'Added quotes for {len(this_matched)} compounds')

			matched += this_matched
			unmatched -= this_matched

			return data
		
			# break

		logger.var('Total time = ', time.perf_counter() - t_start)

		if unmatched:
			logger.error(f'Did not find quotes for {len(unmatched)} compounds')

		if errors:
			logger.error(f'{len(errors)} failed requests')

		return dict(matched=matched, unmatched=unmatched, errors=errors)

	def parse_enamine_bulk_products(self, db, products):

		from .tools import flat_inchikey
		from .cset import CompoundSet

		matched = CompoundSet(db)

		for product in products:

			prices = product['prices']
			product = product['product']

			# quote-level data from product
			entry = product['code']
			smiles = product['smile']
			currency = prices['currency']
			catalogue = ENAMINE_V2_CATALOGUES[product['catalog']]
			
			# check if the product has a name
			if not entry:
				logger.warning(f'Product in {catalogue} but no entry-name. {smiles=}')
				
			if p := product['purity']:
				purity = product['purity'] / 100
			else:
				purity = None

			assert catalogue, product
			assert currency, product

			# identify the compound
			compound_id = db.get_compound_id(inchikey=flat_inchikey(smiles))
			matched.add(compound_id)
			
			# loop through prices
			for pack in prices['all']:

				assert pack['weight']['measure'] == 'mg'
				amount = pack['weight']['amount']
				price = pack['price']

				availability_str = pack['weight']['available']

				try:
					lead_time = ENAMINE_V2_LEAD_TIME[catalogue][availability_str]
				except KeyError:
					logger.error(f'Unsupported {availability_str=} [{compound_id=}, {entry=}]')
					print(pack)
					continue

				# register the quote
				quote_id = db.insert_quote(
					compound=compound_id, 
					supplier='Enamine', 
					catalogue=catalogue, 
					entry=entry, 
					amount=amount, 
					price=price, 
					currency=currency, 
					purity=purity, 
					lead_time=lead_time, 
					smiles=smiles,
					commit=False,
				)
			
			# break

		logger.debug('Committing changes to the database...')
		db.commit()
		
		return matched

	def get_enamine_quote(self, compound, currency="USD"):

		assert compound._table == 'compound'

		try:
		
			smiles = compound.smiles

			logger.header(f'Exact search: {smiles=}')

			for catalogue in self.catalogues:

				logger.header(f'Searching in {self.supplier} {catalogue}...')

				try:
					result = self.wrapper.exactsearch(smiles=compound.smiles, currency=currency, catalogue=catalogue)
				except requests.ConnectionError as e:
					logger.error(f'ConnectionError: {e}')
					return None

				# try again but with a similarity search
				if result['response']['result']['code'] != 0:
					try:
						result = self.wrapper.similaritysearch(smiles, similarity_value=1.0, catalogue=catalogue)
					except requests.ConnectionError as e:
						logger.error(f'ConnectionError: {e}')
						return None

				if result['response']['result']['code'] != 0:
					continue

				data = result['response']['data']

				if len(data) < 1:
					print(result['response'])
					raise NotImplementedError

				data = data[0]

				name = data['Id']

				logger.success(f'Found in {catalogue} w/ id={name}')

				break

			else:

				logger.error('No match found')
				return None
			
			### get the information
				
			logger.header(f'Searching by ID in {self.supplier} {catalogue}...')

			for catalogue in self.catalogues:
				
				result = self.wrapper.compoundidsearch(name, catalogue=catalogue)

				if result['response']['result']['code'] == 0:
					try:
						logger.success(f'Found in {catalogue}')
						return self.parse_enamine_response(result, compound, catalogue), result
					except NoDataInReponse:
						pass

			else:

				logger.error('No catalog entry')
				return None

		except KeyboardInterrupt:
			logger.warning('Interrupted quoting')

	def pick_enamine_exact_data(self, smiles, data):
	
		if len(data) == 1:
			return data[0]['Id']

		if len(data) == 0:
			raise NoDataInResponse(smiles)

		# sort the results by increasing lead_time, and increasing stereo-complexity
		data = sorted(data, key=lambda x: (self.parse_enamine_delivery_string(x['deliveryDays']), stereo_complexity(x['smile'])))
		mout.warning(f'Picking quickest and least stereo complex result from: {[x["Id"] for x in data]}')
		return data[0]['Id']

	def parse_enamine_response(self, result, compound, catalogue):

		quotes = []

		for i,data in enumerate(result['response']['data']):

			if i == 0:

				# update compound metadata
				metadata = dict()
				metadata['enamine_id'] = data['Id']
				metadata['alias'] = data['name']
				metadata['cas'] = data['cas']
				metadata['mfcd'] = data['mfcd']
				metadata['formula'] = data['formula']
				metadata['molecular_weight'] = data['mw']
				compound.metadata.update(metadata)

				"""
				do something with: 
				* availability
				* storageCond
				* productUrl
				* lastUpdate
				"""
			
			lead_time = self.parse_enamine_delivery_string(delivery := data['deliveryDays'])
			entry = data['Id']
			smiles = data['smile']

			if purity := data['purity']:
				purity /= 100
			
			# if not (packs := data['packs']):

			if not (packs := data['packs']) and 'synthesis' in delivery:
				
				logger.warning('Hardcoded REAL quote')

				compound.db.insert_quote(
					compound=compound,
					supplier='Enamine',
					catalogue=catalogue,
					entry=entry,
					amount=20,
					price=330,
					currency='USD',
					purity=None,
					lead_time=15,
					smiles=smiles,
				)

			else:
				for pack in packs:
					self.parse_enamine_pack(compound, entry, purity, catalogue, pack, lead_time, smiles)

	def pick_enamine_data(self, comp_id,data):
		
		data = [d for d in data if d['Id'] == comp_id]
		
		if len(data) == 0:
			raise NoDataInReponse

		elif len(data) == 1:
			return data[0]

		data = [d for d in data if not any([p['price'] == 0.0 for p in d['packs']])]

		if len(data) > 1:
			logger.warning(f'Taking first data entry after stripping non-priced ({comp_id})')
		return data[0]

	def parse_enamine_delivery_string(self, string):

		if string.startswith('regular delivery, ') and string.endswith('days'):
			return int(string.removeprefix('regular delivery, ').removesuffix(' days'))

		if string.startswith('backorder, ') and string.endswith('days'):
			return int(string.removeprefix('backorder, ').removesuffix(' days'))

		if string == '3 weeks synthesis time':
			return 15

		raise Exception(f'Unexpected delivery string: {string}')

	def parse_enamine_pack(self,compound, entry, purity, catalogue, pack, lead_time, smiles):

		supplier = 'Enamine'

		match pack['measure']:
			case 'g':
				amount = pack['amount'] * 1000
			case 'mg':
				amount = pack['amount']
			case _:
				raise NotImplementedError(f'Not supported pack measure: {pack["measure"]}')

		price = pack['price']
		currency = pack['currencyName']

		if pack['status'] not in ['Normal', 'Ice pack', '']:
			print(pack)
			logger.warning(f"{pack['status']=}")

		if not price:
			logger.warning(f'Skipping price-less Enamine pack ({entry})')
			return None

		compound.db.insert_quote(
			compound=compound,
			supplier=supplier,
			catalogue=catalogue,
			entry=entry,
			amount=amount,
			price=price,
			currency=currency,
			purity=purity,
			lead_time=lead_time,
			smiles=smiles,
		)

	### MCULE

	def get_mcule_quote(self, compound, exact=False):

		try:

			# single exact query

			smiles = compound.smiles

			if exact:
				logger.header(f'Exact search: {smiles=}')
				result = self.wrapper.singlequerysearch(smiles)
			else:
				logger.header(f'Similarity==1.0 search: {smiles=}')
				result = self.wrapper.similaritysearch(smiles, threshold=1.0)

			if result['response']['results']:

				results = result['response']['results']

				logger.header(results)

				mcule_ids = [data['mcule_id'] for data in results]
				smiles = [data['smiles'] for data in results]

				logger.success(f'Found w/ ids={mcule_ids}')

				for mcule_id,smile in zip(mcule_ids, smiles):

					# logger.header(f'{self.wrapper.compoundavailability(mcule_id)=}')
					# logger.header(f'{self.wrapper.compounddetails(mcule_id)=}')
					result = self.wrapper.compoundprices(mcule_id)

					for pack in result['response']['best_prices']:
						self.parse_mcule_pack(compound.db, compound, mcule_id, pack, smile)

					# logger.header(f'{self.wrapper.compoundpricesamount(mcule_id)=}')

			return result

		except KeyboardInterrupt:
			logger.warning('Interrupted quoting')

	def get_mcule_batch_quote(self, compounds):

		import time
		from .cset import CompoundSet
		from .tools import flat_inchikey
		from tqdm import tqdm

		db = compounds.db
		
		n_comps = len(compounds)

		matched = CompoundSet(db)

		logger.var('batch size', n_comps)
		assert n_comps <= 1000

		# bulk query compound availability

		t_start = time.perf_counter()
		logger.title('MCule bulk availability query...')
		data = self.wrapper.multiplequerieswithavailability(compounds.smiles)
		
		logger.var('availability query time', time.perf_counter() - t_start)

		results = data['response']['results']
		
		logger.var('#results', len(results))

		logger.title('MCule single price queries...')
		for result in tqdm(results):
			
			entry = result['mcule_id']
			smiles = result['smiles']

			# match to compound
			compound_id = db.get_compound_id(inchikey=flat_inchikey(smiles))

			if not compound_id:
				logger.error(f'Could not find compound matching {smiles=}')
				continue
			
			matched.add(compound_id)

			prices = self.wrapper.compoundprices(entry)
			
			try:
				best_prices = prices['response']['best_prices']
			except KeyError as e:
				print(prices)
				logger.error(f'Unsupported prices {e}')
				continue

			for pack in prices['response']['best_prices']:
				try:
					self.parse_mcule_pack(db, compound_id, entry, pack, smiles, commit=False)
				except KeyError as e:
					print(pack)
					logger.error(f'Unsupported pack {e}')
					continue

		logger.var('Total time = ', time.perf_counter() - t_start)

		unmatched = compounds - matched

		if unmatched:
			logger.error(f'Did not find quotes for {len(unmatched)} compounds')

		logger.success('Committing changes to the database...')
		db.commit()

		return dict(matched=matched, unmatched=unmatched)
	
	def parse_mcule_pack(self, db, compound, entry, pack, smiles, commit=True):

		supplier = 'MCule'

		match pack['unit']:
			case 'g':
				amount = pack['amount'] * 1000
			case 'mg':
				amount = pack['amount']
			case _:
				raise NotImplementedError(f'Not supported pack measure: {pack["unit"]}')

		price = pack['price']
		currency = pack['currency']

		if 'purity' in pack:
			purity = pack['purity'] / 100
		else:
			purity = None
			
		lead_time = pack['delivery_time_working_days']

		db.insert_quote(
			compound=compound,
			supplier=supplier,
			catalogue=None,
			entry=entry,
			amount=amount,
			price=price,
			currency=currency,
			purity=purity,
			lead_time=lead_time,
			smiles=smiles,
			commit=commit,
		)

class NoDataInReponse(Exception):
	...

class NotInCatalogues(Exception):
	...




# from .tools import df_row_to_dict

from .db import Database
from .pose import Pose
from .cset import IngredientSet

import mcol

import os

import logging
logger = logging.getLogger('HIPPO')

class PoseTable:
	"""Object representing the 'pose' table in the :class:`.Database`."""

	def __init__(self,
		db: Database,
		table: str = 'pose',
	) -> None:

		self._db = db
		self._table = table
		
	### FACTORIES

	### PROPERTIES
	
	@property
	def db(self) -> Database:
		"""Returns the associated :class:`.Database`"""
		return self._db
	
	@property
	def table(self) -> str:
		return self._table

	@property
	def names(self):
		"""Returns the aliases of child poses"""
		return [p.name for p in self]

	@property
	def aliases(self):
		"""Returns the aliases of child poses"""
		result = self.db.select(table=self.table, query='pose_alias', multiple=True)
		return [q for q, in result]

	@property
	def inchikeys(self):
		"""Returns the inchikeys of child poses"""
		result = self.db.select(table=self.table, query='pose_inchikey', multiple=True)
		return [q for q, in result]

	@property
	def ids(self):
		"""Returns the IDs of child poses"""
		result = self.db.select(table=self.table, query='pose_id', multiple=True)
		return [q for q, in result]

	@property
	def tags(self):
		"""Returns the set of unique tags present in this pose set"""
		values = self.db.select_where(table='tag', query='DISTINCT tag_name', key='tag_pose IS NOT NULL', multiple=True)
		return set(v for v, in values)

	### METHODS

	def get_by_tag(self, tag):
		"""Get all child poses with a certain tag"""
		values = self.db.select_where(query='tag_pose', table='tag', key='name', value=tag, multiple=True)
		ids = [v for v, in values if v]
		# print(values)
		return self[ids]

	def get_by_target(self, *, id):
		assert isinstance(id, int)
		values = self.db.select_where(query='pose_id', table='pose', key='target', value=id, multiple=True)
		ids = [v for v, in values if v]
		return self[ids]

	def get_by_metadata(self, key: str, value: str | None = None):
		"""Get all child podrd with by their metadata. If no value is passed, then simply containing the key in the metadata dictionary is sufficient"""
		results = self.db.select(query='pose_id, pose_metadata', table='pose', multiple=True)
		if value is None:
			ids = [i for i,d in results if d and f'"{key}":' in d]
		else:
			if isinstance(value, str):
				value = f'"{value}"'
			ids = [i for i,d in results if d and f'"{key}": {value}' in d]
		return self[ids]	

	def draw(self, max_draw=100):
		if len(self) <= max_draw:
			self[:].draw()
		else:
			logger.warning(f"Too many poses: {len(self)} > {max_draw=}. Increase max_draw or use animal.poses[:].draw()")		

	def summary(self):
		"""Print a summary of this pose set"""
		logger.header('PoseTable()')
		logger.var('#poses', len(self))
		logger.var('tags', self.tags)

	def interactive(self):
		return self[self.ids].interactive()

	### DUNDERS

	def __call__(self, tag=None, target=None):
		if tag:
			return self.get_by_tag(tag)
		elif target:
			return self.get_by_target(id=target)
		else:
			raise NotImplementedError

	def __getitem__(self, key) -> Pose:
		
		match key:

			case int():
				if key == 0:
					return self.__getitem__(key=1)

				if key < 0:
					key = len(self) + 1 + key
					return self.__getitem__(key=key)

				else:
					return self.db.get_pose(table=self.table, id=key)

			case str():
				pose = self.db.get_pose(alias=key)
				if not pose:
					pose = self.db.get_pose(inchikey=key)
				return pose				

			case key if isinstance(key, list) or isinstance(key, tuple) or isinstance(key, set):

				indices = []
				for i in key:
					if isinstance(i,int):
						index = i
					elif isinstance(i,str):
						index = self.db.get_pose_id(alias=i)
						if not index:
							index = self.db.get_pose_id(inchikey=i)
					else:
						raise NotImplementedError

					assert index
					indices.append(index)

				return PoseSet(self.db, indices)

			case slice():
				ids = self.db.slice_ids(table=self.table, start=key.start, stop=key.stop, step=key.step)
				return self[ids]

			case _:
				logger.error(f'Unsupported type for PoseTable.__getitem__(): {type(key)}')

		return None

	def __repr__(self) -> str:
		return f'{mcol.bold}{mcol.underline}'"{"f'P x {len(self)}'"}"f'{mcol.unbold}{mcol.ununderline}'

	def __len__(self) -> int:
		return self.db.count(self.table)

	def __iter__(self):
		return iter(self[i+1] for i in range(len(self)))


class PoseSet:
	"""Object representing a subset of the 'pose' table in the :class:`.Database`."""

	_table = 'pose'
	
	def __init__(self,
		db: Database,
		indices: list = None,
	):
		self._db = db

		indices = indices or []

		if not isinstance(indices, list):
			indices = list(indices)

		assert all(isinstance(i, int) for i in indices)

		self._indices = sorted(list(set(indices)))

	### PROPERTIES

	@property
	def db(self):
		return self._db

	@property
	def table(self):
		return self._table

	@property
	def indices(self):
		"""Returns the ids of poses in this set"""
		return self._indices

	@property
	def ids(self):
		"""Returns the ids of poses in this set"""
		return self._indices

	@property
	def names(self):
		"""Returns the aliases of poses in this set"""
		return [p.name for p in self]

	@property
	def aliases(self):
		"""Returns the aliases of child poses"""
		return [self.db.select_where(table=self.table, query='pose_alias', key='id', value=i, multiple=False)[0] for i in self.indices]

	@property
	def inchikeys(self):
		"""Returns the inchikeys of child poses"""
		return [self.db.select_where(table=self.table, query='pose_inchikey', key='id', value=i, multiple=False)[0] for i in self.indices]

	@property
	def smiles(self):
		"""Returns the smiles of poses in this set"""
		return [self.db.select_where(table=self.table, query='pose_smiles', key='id', value=i, multiple=False)[0] for i in self.indices]

	@property
	def tags(self):
		"""Returns the set of unique tags present in this pose set"""
		values = self.db.select_where(table='tag', query='DISTINCT tag_name', key=f'tag_pose in {tuple(self.ids)}', multiple=True)
		return set(v for v, in values)

	@property
	def compounds(self):
		"""Get the compounds associated to this set of poses"""
		from .cset import CompoundSet
		ids = self.db.select_where(table='pose', query='DISTINCT pose_compound', key=f'pose_id in {tuple(self.ids)}', multiple=True)
		ids = [v for v, in ids]
		return CompoundSet(self.db, ids)

	@property
	def mols(self):
		"""Get the rdkit Molecules contained in this set"""
		return [p.mol for p in self]

	@property
	def num_compounds(self):
		"""Count the compounds associated to this set of poses"""
		return len(self.compounds)

	@property
	def df(self):
		"""Get a DataFrame of the poses in this set"""
		return self.get_df(mol=True)

	@property
	def reference_ids(self):
		values = self.db.select_where(table='pose', query='DISTINCT pose_reference', key=f'pose_reference IS NOT NULL and pose_id in {tuple(self.ids)}', value=None, multiple=True)
		return set(v for v, in values)

	# @property
	# def inspirations(self):
	# 	return self._inspirations

	@property
	def inspiration_sets(self):
		sets = []
		for id in self.ids:
			insp_ids = self.db.select_where(query='inspiration_original', table='inspiration', key='derivative', value=id, multiple=True, sort='inspiration_original')
			insp_ids = set(v for v, in insp_ids)
			# print(insp_ids)
			if insp_ids not in sets:
				sets.append(insp_ids)
		return sets

	@property
	def str_ids(self):
		return str(tuple(self.ids)).replace(',)',')')

	@property
	def targets(self):
		"""Returns the targets of poses in this set"""
		return [self.db.get_target(id=q) for q in self.target_ids]

	@property
	def target_names(self):
		"""Returns the targets of poses in this set"""
		return [self.db.get_target_name(id=q) for q in self.target_ids]

	@property
	def target_ids(self):
		"""Returns the target ID's of poses in this set"""
		result = self.db.select_where(table=self.table, query='DISTINCT pose_target', key=f'pose_id in {self.str_ids}', multiple=True)
		return [q for q, in result]

	@property
	def best_placed_pose(self):
		"""Returns the pose with the best distance_score in this subset"""
		return self.db.get_pose(id=self.best_placed_pose_id)

	@property
	def best_placed_pose_id(self):
		query = f'pose_id, MIN(pose_distance_score)'
		query = self.db.select_where(table='pose', query=query, key=f'pose_id in {self.str_ids}', multiple=False)
		return query[0]

	### FILTERING

	def get_by_tag(self, tag, inverse=False):
		"""Get all child poses with a certain tag"""
		values = self.db.select_where(query='tag_pose', table='tag', key='name', value=tag, multiple=True)
		if inverse:
			matches = [v for v, in values if v]
			ids = [i for i in self.ids if i not in matches]
		else:
			ids = [v for v, in values if v and v in self.ids]
		return PoseSet(self.db, ids)

	def get_by_metadata(self, key: str, value: str | None = None):
		"""Get all child poses with by their metadata. If no value is passed, then simply containing the key in the metadata dictionary is sufficient"""
		results = self.db.select(query='pose_id, pose_metadata', table='pose', multiple=True)
		if value is None:
			ids = [i for i,d in results if d and f'"{key}":' in d and i in self.ids]
		else:
			if isinstance(value, str):
				value = f'"{value}"'
			ids = [i for i,d in results if d and f'"{key}": {value}' in d and i in self.ids]
		return PoseSet(self.db, ids)		

	def get_by_inspiration(self, inspiration: int | Pose, inverse=False):
		"""Get all child poses with with this inspiration."""

		ids = set()

		for pose in self:
			if not inverse:
				for pose_inspiration in pose.inspirations:
					if pose_inspiration == inspiration:
						ids.add(pose.id)
						break

			elif inverse:
				for pose_inspiration in pose.inspirations:
					if pose_inspiration == inspiration:
						break
				else:
					ids.add(pose.id)

		return PoseSet(self.db, ids)

	def get_df(self, skip_no_mol=True, **kwargs):
		"""Get a DataFrame of the poses in this set"""
		
		from pandas import DataFrame

		data = []

		for pose in self:
			d = pose.get_dict(**kwargs)

			if skip_no_mol and not d['mol']:
				logger.warning(f'Skipping pose with no mol: {d["id"]} {d["name"]}')
				continue
			data.append(d)

		return DataFrame(data)

	def get_by_reference(self, ref_id):
		"""Get poses with a certain reference id"""
		values = self.db.select_where(table='pose', query='pose_id', key='reference', value=ref_id, multiple=True)
		if not values:
			return None
		return PoseSet(self.db, [v for v, in values])

	def get_by_target(self, *, id):
		assert isinstance(id, int)
		values = self.db.select_where(query='pose_id', table='pose', key=f'pose_target is {id} AND pose_id in {self.str_ids}', multiple=True, none='quiet')
		ids = [v for v, in values if v]
		if not ids:
			return None
		return PoseSet(self.db, ids)

	def filter(self, function, inverse=False):
		"""Filter this poseset by selecting members where function(pose) is truthy"""
		
		ids = set()
		for pose in self:
			value = function(pose)
			# logger.debug(f'{pose=} {value=}')
			if value and not inverse:
				ids.add(pose.id)
			elif not value and inverse:
				ids.add(pose.id)

		return PoseSet(self.db, ids)

	### BULK SETTING

	@property
	def reference(self):
		raise NotImplementedError
	
	@reference.setter
	def reference(self, r):
		"""Bulk set the references for poses in this set"""
		if not isinstance(r, int):
			assert r._table == 'pose'
			r = r.id

		for i in self.indices:
			self.db.update(table='pose', id=i, key='pose_reference', value=r, commit=False)

		self.db.commit()
	
	def add_tag(self, t):
		"""Add this tag to every member of the set"""

		assert isinstance(t, str)

		for i in self.indices:
			self.db.insert_tag(name=t, pose=i, commit=False)	

		logger.info(f'Tagged {self} w/ "{t}"')		

		self.db.commit()

	def append_to_metadata(self, key, value):
		"""Create or append to a list-like value with given key for each pose in this set"""
		for id in self.indices:
			metadata = self.db.get_metadata(table='pose', id=id)
			metadata.append(key, value)

	### SPLITTING

	def split_by_reference(self):
		sets = {}
		for ref_id in self.reference_ids:
			sets[ref_id] = self.get_by_reference(ref_id)
		return sets

	def split_by_inspirations(self, single_set=False):
		
		sets = {}

		for pose in self:

			insp_ids = tuple(pose.get_inspiration_ids())

			if insp_ids not in sets:
				sets[insp_ids] = PoseSet(self.db, [pose.id])
			else:
				sets[insp_ids]._indices.append(pose.id)

		if single_set:
			logger.var('#unique inspiration combinations', len(sets))
			sets = PoseSet(self.db, sum([s.ids for s in sets.values()], []))

		return sets


	### EXPORTING

	def write_sdf(self, out_path, name_col='name'):
		"""Write an SDF"""

		from pathlib import Path

		df = self.get_df(mol=True, inspirations='fragalysis')

		df.rename(inplace=True, columns={name_col:'_Name', 'mol':'ROMol'})

		logger.writing(out_path)
			
		from rdkit.Chem import PandasTools
		PandasTools.WriteSDF(df, out_path, "ROMol", "_Name", list(df.columns))

		# keep record of export
		value = str(Path(out_path).resolve())
		self.db.remove_metadata_list_item(table='pose', key='exports', value=value)
		self.append_to_metadata(key='exports', value=value)

	def to_fragalysis(self, 
		out_path,
		*,
		method,
		ref_url,
		submitter_name,
		submitter_email,
		submitter_institution,
		metadata: bool = True,
		sort_by: str | None = None,
		sort_reverse: bool = False,
		generate_pdbs: bool = False,
		ingredients: IngredientSet = None,
	):

		"""Prepare an SDF for upload to the RHS of Fragalysis"""

		from .fragalysis import generate_header
		from pathlib import Path
		from rdkit.Chem import SDWriter, PandasTools
		assert out_path.endswith('.sdf')

		name_col = '_Name'
		mol_col = 'ROMol'

		# get the dataframe of poses

		pose_df = self.get_df(mol=True, inspirations='fragalysis', duplicate_name='original ID', reference='name', metadata=metadata)
		
		drops = ['path', 'compound', 'target', 'ref_pdb', 'original SMILES']

		if ingredients:
			drops.pop(drops.index('compound'))
		
		pose_df = pose_df.drop(columns=drops, errors='ignore')

		pose_df.rename(inplace=True, columns=
			{'name':name_col, 
			 'id':'HIPPO ID', 
			 'mol':mol_col, 
			 'inspirations':'ref_mols',
			 'reference':'ref_pdb',
			 'smiles':'original SMILES',
			 'compound':'compound inchikey',
			})

		extras={'smiles':'smiles', 'ref_mols':'fragment inspirations', 'original ID':'original ID'}

		if ingredients:

			q_entries = []
			q_prices = []
			q_lead_times = []
			q_amounts = []

			currency = None

			for i,row in pose_df.iterrows():

				compound_id = self.db.get_compound_id(inchikey=row['compound inchikey'])

				ingredient = ingredients(compound_id=compound_id)

				if isinstance(ingredient, IngredientSet):
					ingredient = sorted([i for i in ingredient], key=lambda x: x.quote.price)[0]

				quote = ingredient.quote
				if not currency:
					currency = quote.currency
				else:
					assert quote.currency == currency

				q_entries.append(quote.entry_str)
				q_prices.append(quote.price)
				q_lead_times.append(quote.lead_time)
				q_amounts.append(quote.amount)

			pose_df['Supplier Catalogue Entry'] = q_entries
			# pose_df['Supplier:Catalogue:Entry'] = q_entries
			pose_df[f'Price ({currency})'] = q_prices
			pose_df['Lead time (working days)'] = q_lead_times
			pose_df['Amount (mg)'] = q_amounts

			extras['Supplier Catalogue Entry'] = "Supplier Catalogue Entry string"
			extras[f'Price ({currency})'] = "Quoted price"
			extras['Lead time (working days)'] = "Quoted lead-time"
			extras['Amount (mg)'] = "Quoted amount"

		if generate_pdbs:
			
			from zipfile import ZipFile

			# output subdirectory
			out_key = Path(out_path).name.removesuffix('.sdf')
			pdb_dir = Path(out_key)
			pdb_dir.mkdir(exist_ok=True)

			# create the zip archive
			with ZipFile(f'{out_key}_pdbs.zip', 'w') as z:
				
				# loop over poses
				for (i,row),pose in zip(pose_df.iterrows(), self):

					# filenames
					pdb_name = f"{out_key}_{row._Name}.pdb"
					pdb_path = pdb_dir / pdb_name
					pose_df.loc[i, 'ref_pdb'] = pdb_name

					# generate the PL-complex
					sys = pose.complex_system
					
					# write the PDB
					logger.writing(pdb_path)
					sys.write(pdb_path, verbosity=0)
					z.write(pdb_path)
			
			logger.writing(f'{out_key}_pdbs.zip')

		# create the header molecule
		
		df_cols = set(pose_df.columns)

		header = generate_header(
			self[0],
			method=method,
			ref_url=ref_url,
			submitter_name=submitter_name,
			submitter_email=submitter_email,
			submitter_institution=submitter_institution,
			extras=extras,
			metadata=metadata,
		)

		header_cols = set(header.GetPropNames())

		# empty properties
		pose_df['generation_date'] = [None] * len(pose_df)
		pose_df['submitter_name'] = [None] * len(pose_df)
		pose_df['method'] = [None] * len(pose_df)
		pose_df['submitter_email'] = [None] * len(pose_df)
		pose_df['ref_url'] = [None] * len(pose_df)

		if sort_by:
			pose_df = pose_df.sort_values(by=sort_by, ascending=not sort_reverse)

		fields = []

		logger.writing(out_path)

		with open(out_path, 'w') as sdfh:
			with SDWriter(sdfh) as w:
				w.write(header)
			PandasTools.WriteSDF(pose_df, sdfh, mol_col, name_col, set(pose_df.columns))

		# keep record of export
		value = str(Path(out_path).resolve())
		self.db.remove_metadata_list_item(table='pose', key='exports', value=value)
		self.append_to_metadata(key='exports', value=value)

		return pose_df

	def to_pymol(self, prefix=None):
		"""Group the poses by reference protein and inspirations and output relevant PDBs and SDFs."""

		commands = []

		prefix = prefix or ''
		if prefix:
			prefix = f'{prefix}_'

		from pathlib import Path

		for i,(ref_id, poses) in enumerate(self.split_by_reference().items()):

			ref_pose = self.db.get_pose(id=ref_id)
			ref_name = ref_pose.name or ref_id
		
			# create the subdirectory
			ref_dir = Path(f'{prefix}ref_{ref_name}')
			logger.writing(ref_dir)
			ref_dir.mkdir(parents=True, exist_ok=True)
			
			# write the reference protein
			ref_pdb = ref_dir / f'ref_{ref_name}.pdb'
			ref_pose.protein_system.write(ref_pdb, verbosity=0)

			# color the reference:
			commands.append(f'load {ref_pdb.resolve()}')
			commands.append('hide')
			commands.append('show lines')
			commands.append('show surface')
			commands.append('util.cbaw')
			commands.append('set surface_color, white')
			commands.append('set transparency,  0.4')

			for j,(insp_ids, poses) in enumerate(poses.split_by_inspirations().items()):

				inspirations = PoseSet(self.db, insp_ids)
				insp_names = "-".join(inspirations.names)

				# create the subdirectory
				insp_dir = ref_dir / insp_names
				insp_dir.mkdir(parents=True, exist_ok=True)

				# write the inspirations
				insp_sdf = insp_dir / f'{insp_names}_frags.sdf'
				inspirations.write_sdf(insp_sdf)

				commands.append(f"load {insp_sdf.resolve()}")
				commands.append(f"set all_states, on, {insp_sdf.name.removesuffix('.sdf')}")
				commands.append(f"util.rainbow \"{insp_sdf.name.removesuffix('.sdf')}\"")

				# write the poses
				pose_sdf = insp_dir / f'{insp_names}_derivatives.sdf'
				poses.write_sdf(pose_sdf)
				
				commands.append(f"load {pose_sdf.resolve()}")
				commands.append(f'util.cbaw "{pose_sdf.name.removesuffix(".sdf")}"')

				if j > 0:
					commands.append(f"disable \"{insp_sdf.name.removesuffix('.sdf')}\"")
					commands.append(f'disable "{pose_sdf.name.removesuffix(".sdf")}"')

		return '; '.join(commands)

	### OUTPUT

	def interactive(self, method=None, print_name=True, **kwargs):
		"""Creates a ipywidget to interactively navigate this PoseSet."""

		from ipywidgets import interactive, BoundedIntText, Checkbox, interactive_output, HBox, GridBox, Layout, VBox
		from IPython.display import display
		from pprint import pprint

		if method:
			def widget(i):
				pose = self[i]
				if print_name:
					print(repr(pose))
				value = getattr(pose, method)(**kwargs)
				if value:
					display(value)

			return interactive(widget, i=
				BoundedIntText(
					value=0,
					min=0,
					max=len(self)-1,
					step=1,
					description='Pose:',
					disabled=False
				))

		else:

			a = BoundedIntText(
					value=0,
					min=0,
					max=len(self)-1,
					step=1,
					description=f'Pose (/{len(self)}):',
					disabled=False,
				)

			b = Checkbox(description='Name', value=True)
			c = Checkbox(description='Summary', value=False)
			d = Checkbox(description='2D', value=False)
			e = Checkbox(description='3D', value=True)
			f = Checkbox(description='Metadata', value=False)

			ui = GridBox([b, c, d, e, f], layout=Layout(grid_template_columns="repeat(5, 100px)"))
			ui = VBox([a, ui])
			
			def widget(i, name=True, summary=True, grid=True, draw=True, metadata=True):
				pose = self[i]
				if name:
					print(repr(pose))

				if summary: pose.summary(metadata=False)
				if grid: pose.grid()
				if draw: pose.draw()
				if metadata:
					logger.title('Metadata:')
					pprint(pose.metadata)

			out = interactive_output(widget, {'i': a, 'name': b, 'summary': c, 'grid':d, 'draw':e, 'metadata':f})

			display(ui, out)

	def summary(self):
		"""Print a summary of this pose set"""
		logger.header('PoseSet()')
		logger.var('#poses', len(self))
		logger.var('#compounds', self.num_compounds)
		logger.var('tags', self.tags)

	def draw(self):
		"""Render this pose set with Py3Dmol"""
		
		from molparse.rdkit import draw_mols

		mols = [p.mol for p in self]

		drawing = draw_mols(mols)
		display(drawing)

	def grid(self):
		"""Draw a grid of all contained molecules"""
		from molparse.rdkit import draw_grid

		data = [(p.name, p.compound.mol) for p in self]

		mols = [d[1] for d in data]
		labels = [d[0] for d in data]

		return draw_grid(mols, labels=labels)

	### DUNDERS

	def __repr__(self) -> str:
		return f'{mcol.bold}{mcol.underline}'"{"f'P x {len(self)}'"}"f'{mcol.unbold}{mcol.ununderline}'

	def __len__(self) -> int:
		return len(self.indices)

	def __iter__(self):
		return iter(self.db.get_pose(table=self.table, id=i) for i in self.indices)

	def __getitem__(self, key) -> Pose:
		match key:
			
			case int():
				try:
					index = self.indices[key]
				except IndexError:
					logger.exception(f'list index out of range: {key=} for {self}')
					raise
				return self.db.get_pose(table=self.table, id=index)
			
			case slice():
				ids = self.indices[key]
				return PoseSet(self.db, ids)

			case _:
				raise NotImplementedError





import mcol
import molparse as mp

from rdkit import Chem

import numpy as np

from .tags import TagSet

import pickle

from pathlib import Path

import logging
logger = logging.getLogger('HIPPO')

from molparse.rdkit.features import FEATURE_FAMILIES, COMPLEMENTARY_FEATURES

CUTOFF_PADDING = 1.0

FEATURE_PAIR_CUTOFFS = {
	'Donor Acceptor': 3.5 + CUTOFF_PADDING,
	'Acceptor Donor': 3.5 + CUTOFF_PADDING,
	'NegIonizable PosIonizable': 4.5 + CUTOFF_PADDING,
	'PosIonizable NegIonizable': 4.5 + CUTOFF_PADDING,
	'Aromatic PosIonizable': 4.5 + CUTOFF_PADDING,
	'PosIonizable Aromatic': 4.5 + CUTOFF_PADDING,
	'Aromatic Aromatic': 6.0 + CUTOFF_PADDING,
	'Hydrophobe Hydrophobe': 4.5 + CUTOFF_PADDING,
}

class Pose:

	"""A :class:`.Pose` is a particular conformer of a :class:`.Compound` within a protein environment. A pose will have its own (stereochemical) smiles string, and must have a path to a coordinate file. Poses can have *inspirations* that can be used to trace fragment-derived scaffolds in merges and expansions.

	:class:`.Pose` objects should not be created directly. Instead use :meth:`.HIPPO.register_pose` or :meth:`.HIPPO.poses`
	"""

	_table = 'pose'

	def __init__(self, 
		db,
		id: int,
		inchikey: str | None,
		alias: str | None,
		smiles: str,
		reference: int, # another pose
		path: str,
		compound: int,
		target: str,
		mol: Chem.Mol | bytes | None,
		fingerprint: bytes,
		energy_score: float | None = None,
		distance_score: float | None = None,
		metadata: dict | None = None,
	):

		self._db = db
		self._id = id
		self._inchikey = inchikey
		self._alias = alias
		self._smiles = smiles
		self._compound_id = compound
		self._target = target
		self._path = path
		self._protein_system = None
		self._energy_score = energy_score
		self._distance_score = distance_score

		self._base_id = None
		self._num_heavy_atoms = None

		if fingerprint:
			# logger.debug('unpickling fingerprint')
			fingerprint = pickle.loads(fingerprint)		
		self._fingerprint = fingerprint

		# print(f'{self}{metadata=}')
		self._metadata = metadata
		self._tags = None
		self._reference = reference

		if isinstance(mol, bytes):
			self._mol = Chem.Mol(mol)
		else: 
			self._mol = mol
		
	### FACTORIES

	### PROPERTIES

	@property
	def db(self):
		"""Returns a pointer to the parent database"""
		return self._db

	@property
	def id(self) -> int:
		"""Returns the pose's database ID"""
		return self._id

	@property
	def inchikey(self) -> str:
		"""Returns the pose's inchikey"""
		if not self._inchikey:
			self.smiles
		return self._inchikey

	@property
	def alias(self) -> str:
		"""Returns the pose's alias"""
		return self._alias

	@property
	def name(self) -> str:
		"""Returns the pose's name"""
		if n := self.alias:
			return n
		else:
			return self.inchikey

	@alias.setter
	def alias(self, n):
		"""Set the pose's alias"""
		assert isinstance(n, str)
		self._alias = n
		self.db.update(table='pose', id=self.id, key='pose_alias', value=n)

	@inchikey.setter
	def inchikey(self, n):
		"""Set the pose's inchikey"""
		assert isinstance(n, str)
		self._inchikey = n
		self.db.update(table='pose', id=self.id, key='pose_inchikey', value=n)

	@property
	def smiles(self):
		"""Returns the pose's smiles"""
		if not self._smiles:
			from molparse.rdkit import mol_to_smiles
			from rdkit.Chem.inchi import MolToInchiKey
			try:
				mol = self.mol
				self._smiles = mol_to_smiles(mol)
				self.inchikey = MolToInchiKey(mol)
				self.db.update(table='pose', id=self.id, key='pose_smiles', value=self._smiles)
			except InvalidMolError:
				logger.warning(f'Taking smiles from {self.compound}')
				self._smiles = self.compound.smiles
		return self._smiles

	@property
	def target(self):
		"""Returns the pose's associated target"""
		if isinstance(self._target, int):
			self._target = self.db.get_target(id=self._target)
		return self._target

	@property
	def compound_id(self) -> int:
		"""Returns the pose's associated compound ID"""
		return self._compound_id

	@property
	def compound(self):
		"""Returns the pose's associated compound"""
		return self.get_compound()

	@property
	def path(self):
		"""Returns the pose's path"""
		return self._path

	@property
	def reference(self):
		"""Returns the pose's protein reference (another pose)"""
		if isinstance(self._reference, int):
			self._reference = self.db.get_pose(id=self._reference)
		return self._reference

	@reference.setter
	def reference(self, p):
		"""Set the pose's reference"""
		if not isinstance(p, int):
			assert p._table == 'pose'
			p = p.id
		self._reference = p
		self.db.update(table='pose', id=self.id, key='pose_reference', value=p)

	@property
	def mol(self):
		"""Returns a pose's rdkit.Chem.Mol"""
		if not self._mol and self.path:

			if self.path.endswith('.pdb'):

				# logger.reading(self.path)
				sys = mp.parse(self.path, verbosity=False)
				
				self.protein_system = sys.protein_system

				# look for ligand mol from Fragalysis
				mol_path = list(Path(self.path).parent.glob("*_ligand.mol")) # str(Path(self.path).name).replace('.pdb','_ligand.mol')
				
				if len(mol_path) == 1:
					mol_path = mol_path[0]
					from rdkit.Chem import MolFromMolFile
					mol = MolFromMolFile(str(mol_path.resolve()))

				elif len(mol_path) == 0:
				
					lig_residues = sys['rLIG']

					if not lig_residues:
						lig_residues = [r for r in sys.residues if r.type == 'LIG']

					if len(lig_residues) > 1:
						logger.warning(f'Multiple ligands in PDB {self}')

					lig_res = lig_residues[0]
					
					if not (mol := lig_res.rdkit_mol):
						logger.error(f'[{self}] Error computing RDKit Mol from PDB={self.path}')
						
						print(lig_res.pdb_block)

						lig_res.plot3d()

						raise InvalidMolError

					# clean up bond orders
					from rdkit.Chem.AllChem import MolFromSmiles, AssignBondOrdersFromTemplate
					template = MolFromSmiles(self.compound.smiles)
					try:
						mol = AssignBondOrdersFromTemplate(template, mol)
					except Exception as e:
						logger.error(f'Exception occured during AssignBondOrdersFromTemplate for {self}.mol')
						print(f'template_smiles={self.compound.smiles}')
						print(f'pdbblock={print(lig_res.pdb_block)}')
						logger.error(e)
						mol = lig_res.rdkit_mol
				
				else:

					logger.warning(f'There are multiple *_ligand.mol files in {Path(self.path).parent}')

				self.mol = mol

			elif self.path.endswith('.mol'):

				# logger.reading(self.path)
				mol = mp.parse(self.path, verbosity=False)

				if not mol:
					logger.error(f'[{self}] Error computing RDKit Mol from .mol={self.path}')
					
					raise InvalidMolError

				self.mol = mol

			else:

				raise NotImplementedError

			if not mol:
				logger.error(f'Could not parse {self}.path={self.path}')

		return self._mol

	@mol.setter
	def mol(self, m):
		"""Set the pose's rdkit.Chem.Mol"""
		assert m
		from .tools import sanitise_mol
		self._mol = sanitise_mol(m)
		self.db.update(table='pose', id=self.id, key='pose_mol', value=m.ToBinary())

	@property
	def protein_system(self):
		"""Returns the pose's protein molparse.System"""
		if self._protein_system is None and self.path.endswith('.pdb'):
			# logger.debug(f'getting pose protein system {self}')
			self.protein_system = mp.parse(self.path, verbosity=False).protein_system
		return self._protein_system

	@protein_system.setter
	def protein_system(self, a):
		"""Sets the pose's protein molparse.System"""
		self._protein_system = a

	@property
	def complex_system(self):

		if self.has_complex_pdb_path:
			return mp.parse(self.path, verbosity=False)

		elif self.reference:

			# construct from .mol and reference

			system = self.reference.protein_system.copy()

			system.name = f'{self.target.name}_{self.reference.name}_{self.compound.name}'

			from molparse.rdkit import mol_to_AtomGroup
			ligand = mol_to_AtomGroup(self.mol)

			for atom in ligand.atoms:
				system.add_atom(atom)

			return system

		else:

			raise NotImplementedError

	@property
	def has_complex_pdb_path(self):
		return self.path.endswith('.pdb')

	@property
	def metadata(self):
		"""Returns the pose's metadata"""
		if self._metadata is None:
			self._metadata = self.db.get_metadata(table='pose', id=self.id)
		return self._metadata

	@property
	def fingerprint(self):
		"""Returns the pose's fingerprint"""
		return self._fingerprint

	@fingerprint.setter
	def fingerprint(self, fp):
		"""Set the pose's fingerprint"""

		# remove features that don't exist in this fingerprint?

		self._fingerprint = fp

		# store in the database
		fp = pickle.dumps(fp)
		# logger.debug('pickling fingerprint')
		self.db.update(table='pose', id=self.id, key=f'pose_fingerprint', value=fp)

	@property
	def tags(self):
		"""Returns the pose's tags"""
		if not self._tags:
			self._tags = self.get_tags()
		return self._tags

	@property
	def inspirations(self):
		"""Returns the pose's inspirations"""
		return self.get_inspirations()

	@property
	def features(self):
		"""Returns the pose's features"""
		return mp.rdkit.features_from_mol(self.mol)
	
	@property
	def dict(self):
		"""Serialised dictionary representing the pose"""
		return self.get_dict()

	@property
	def table(self):
		return self._table

	@property
	def num_heavy_atoms(self):
		"""Number of heavy atoms"""
		if not self._num_heavy_atoms:
			# print(self.compound_id)
			self._num_heavy_atoms = self.db.get_compound_computed_property('num_heavy_atoms', self.compound_id)
		return self._num_heavy_atoms

	@property
	def num_atoms_added(self):
		"""Calculate the number of atoms added relative to the base or inspirations"""
		# use base
		if (b_id := self.base_id):
			n_e = self.num_heavy_atoms
			n_b = self.db.get_compound_computed_property('num_heavy_atoms', b_id)
			return n_e - n_b
			
		# use inspirations
		else:
			# raise NotImplementedError
			inspirations = self.inspirations
			assert inspirations
	
			count = 0
			for i in inspirations:
				count += i.num_heavy_atoms
	
			if (self_count := self.num_heavy_atoms) is None:
				return None
			return self.num_heavy_atoms - count

	@property
	def base_id(self):
		"""Get the base compound ID"""
		if not self._base_id:
			val, = self.db.select_where(table='compound', query='compound_base', key='id', value=self.compound_id)
			self._base_id = val
		return self._base_id

	@property
	def fields(self):
		return [p for p in dir(self) if not p.startswith('_')]
    
	@property
	def energy_score(self):
		return self._energy_score

	@property
	def distance_score(self):
		return self._distance_score

	### METHODS

	def get_compound(self):
		return self.db.get_compound(id=self._compound_id)

	def get_tags(self):
		tags = self.db.select_where(query='tag_name', table='tag', key='pose', value=self.id, multiple=True, none='quiet')
		return TagSet(self, {t[0] for t in tags})
	
	def get_inspiration_ids(self):
		inspirations = self.db.select_where(query='inspiration_original', table='inspiration', key='derivative', value=self.id, multiple=True, none='quiet')
		if not inspirations:
			return None
		return set([v for v, in inspirations])

	def get_inspirations(self):	
		if not (inspirations := self.get_inspiration_ids()):
			return None

		from .pset import PoseSet
		return PoseSet(self.db, indices=inspirations)

	def get_dict(self, 
		mol: bool = False, 
		inspirations: bool | str = True, 
		reference: bool | str = True,
		metadata: bool = True,
		duplicate_name: str | bool = False,
	) -> dict:

		"""Returns a dictionary representing this Pose. Arguments:

		mol (bool): [True, False]
		inspirations (bool): [True, False, 'fragalysis']
		reference (bool): [True, False, 'name']

		"""

		serialisable_fields = ['id','inchikey', 'alias', 'name', 'smiles', 'path']

		data = {}
		for key in serialisable_fields:
			data[key] = getattr(self, key)

		if duplicate_name:
			assert isinstance(duplicate_name, str)
			data[duplicate_name] = data['name']

		if mol:
			try:
				data['mol'] = self.mol
			except InvalidMolError:
				data['mol'] = None

		data['compound'] = self.compound.name
		data['target'] = self.target.name

		data['tags'] = self.tags
		
		if inspirations == 'fragalysis':
			data['inspirations'] = ','.join([p.name for p in self.inspirations])
		elif inspirations:
			data['inspirations'] = self.inspirations

		if reference == 'name':
			data['reference'] = self.reference.name
		elif reference:
			data['reference'] = self.reference

		if metadata and (metadict := self.metadata):
			for key in metadict:
				data[key] = metadict[key]

		return data

	def calculate_fingerprint(self):
		"""Calculate the pose's interaction fingerprint"""

		if self.path.endswith('.pdb'):
			
			import molparse as mp
			protein_system = self.protein_system
			if not self.protein_system:
				# logger.reading(self.path)
				protein_system = mp.parse(self.path, verbosity=False).protein_system

		elif self.path.endswith('.mol') and self.reference:

			logger.debug('fingerprint from .mol and reference pose')
			protein_system = self.reference.protein_system

		else:

			logger.debug('calculate_fingerprint()')
			raise NotImplementedError

		assert protein_system

		if not self.mol:
			return

		comp_features = self.features

		comp_features_by_family = {}
		for family in FEATURE_FAMILIES:
			comp_features_by_family[family] = [f for f in comp_features if f.family == family]

		# protein_features = self.target.features
		# if not protein_features:
		protein_features = self.target.calculate_features(protein_system)

		fingerprint = {}

		chains = protein_system.chain_names

		for prot_feature in protein_features:

			if prot_feature.chain_name not in chains:
				continue

			prot_family = prot_feature.family

			prot_residue = protein_system.get_chain(prot_feature.chain_name).residues[f'n{prot_feature.residue_number}']

			if not prot_residue:
				continue

			# if prot_feature.residue_number == 77:
			# 	logger.debug(repr(prot_feature))

			if prot_residue.name != prot_feature.residue_name:
				logger.warning(f'Feature {repr(prot_feature)}')
				continue

			prot_atoms = [prot_residue.get_atom(a).np_pos for a in prot_feature.atom_names.split(' ')]
			
			prot_coord = np.array(np.sum(prot_atoms,axis=0)/len(prot_atoms))

			complementary_family = COMPLEMENTARY_FEATURES[prot_family]

			complementary_comp_features = comp_features_by_family[complementary_family]

			cutoff = FEATURE_PAIR_CUTOFFS[f'{prot_family} {complementary_family}']

			valid_features = [f for f in complementary_comp_features if np.linalg.norm(f - prot_coord) <= cutoff]

			fingerprint[prot_feature.id] = len(valid_features)

		self.fingerprint = fingerprint

	def draw(self):
		"""Render this pose (and its inspirations)"""
		
		from molparse.rdkit import draw_mols
		
		mols = [self.mol]
		if self.inspirations:
			mols += [i.mol for i in self.inspirations]
		
		return draw_mols(mols)

	def grid(self):
		"""Draw a grid of this pose with its inspirations"""
		from molparse.rdkit import draw_grid
		from IPython.display import display

		mols = [self.compound.mol]
		labels = [self.plain_repr()]
		if self.inspirations:
			mols += [i.compound.mol for i in self.inspirations]
			labels += [i.plain_repr() for i in self.inspirations]

		display(draw_grid(mols, labels=labels))

	def summary(self, metadata:bool = True):
		"""Print a summary of this pose"""
		if self.alias:
			logger.header(f'{str(self)}: {self.alias}')
		else:
			logger.header(f'{str(self)}: {self.inchikey}')
		logger.var('inchikey', self.inchikey)
		logger.var('alias', self.alias)
		logger.var('smiles', self.smiles)
		logger.var('compound', repr(self.compound))
		logger.var('path', self.path)
		logger.var('target', repr(self.target))
		logger.var('reference', self.reference)
		logger.var('tags', self.tags)
		logger.var('num_heavy_atoms', self.num_heavy_atoms)
		if (inspirations := self.inspirations):
			logger.var('inspirations', self.inspirations.names)
			logger.var('num_atoms_added', self.num_atoms_added)
		if metadata:
			logger.var('metadata', str(self.metadata))

	def showcase(self):
		self.summary(metadata=False)
		self.grid()
		self.draw()
		from pprint import pprint
		logger.title('Metadata:')
		pprint(self.metadata)

	def plain_repr(self):
		"""Unformatted __repr__"""
		if self.name:
			return f'{self.compound}->{self}: {self.name}'
		else:
			return f'{self.compound}->{self}'

	### DUNDERS

	def __str__(self):
		return f'P{self.id}'

	def __repr__(self):
		return f'{mcol.bold}{mcol.underline}{self.plain_repr()}{mcol.unbold}{mcol.ununderline}'

	def __eq__(self, other):

		if isinstance(other, int):
			return self.id == other

		return self.id == other.id


class InvalidMolError(Exception):
	...



def generate_header(
	pose,
	# metadata,
	method,
	ref_url,
	submitter_name,
	submitter_email,
	submitter_institution,
	generation_date: str | None = None,
	extras=None,
	metadata: bool = True,
):

	extras = extras or {}

	from rdkit.Chem.AllChem import EmbedMolecule
	from molparse.rdkit import mol_from_smiles
	from datetime import date

	header = mol_from_smiles(pose.compound.smiles)

	header.SetProp('_Name', 'ver_1.2')
	EmbedMolecule(header)

	generation_date = str(generation_date or date.today())

	header.SetProp('ref_url', ref_url)
	header.SetProp('submitter_name', submitter_name)
	header.SetProp('submitter_email', submitter_email)
	header.SetProp('submitter_institution', submitter_institution)
	header.SetProp('generation_date', generation_date)
	header.SetProp('method', method)

	if metadata:
		for k,v in pose.metadata.items():
			header.SetProp(k, str(k))

	for k,v in extras.items():
		header.SetProp(k, str(v))

	return header




from mout import debug_log

import logging
logger = logging.getLogger('HIPPO')

"""

Checks
======

- Num heavy atoms difference
- Formula checks
- Num rings difference

"""

SUPPORTED_CHEMISTRY = {

	"Amidation" : 
	{
		"heavy_atoms_diff":1, 
		"rings_diff":0,
		"atomtype":
		{
			"removed" : { "O":1, "H":2 },
		},
	},

	"Williamson_ether_synthesis" : 
	{
		"heavy_atoms_diff":1, 
		"rings_diff":0,
		"atomtype":
		{
			"removed" : { "Ha":1, "H":1 }, # any halogen
		},
	},

	"N-Boc_deprotection" : 
	{
		"heavy_atoms_diff":7, 
		"rings_diff":0,
		"atomtype":
		{
			"removed" : { "O":2, "C":5, "H":8 },
		},
	},

	"TBS_alcohol_deprotection" : 
	{
		"heavy_atoms_diff":7, 
		"rings_diff":0,
		"atomtype":
		{
			"removed" : { "C":6, "Si":1, "H":14 },
		},
	},

	"Sp3-sp2_Suzuki_coupling" : 
	{
		"heavy_atoms_diff":10, 
		"rings_diff":1,
		"atomtype":
		{
			"removed" : { "C":6, "O":2, "B":1, "Ha":1, "H":12 }, # any halogen
		},
	},

	"Benzyl_alcohol_deprotection" : 
	{
		"heavy_atoms_diff":7, 
		"rings_diff":1,
		"atomtype":
		{
			"removed" : { "C":7, "H":6 },
		},
	},

	"Nucleophilic_substitution" : 
	{
		"heavy_atoms_diff":1, 
		"rings_diff":0,
		# "atomtype":
		# {
		# 	"removed" : { "C":7, "H":6 },
		# },
	},

}

def check_reaction_types(types):
	for reaction_type in types:
		if reaction_type not in SUPPORTED_CHEMISTRY:
			logger.warning(f"Can't check chemistry of unsupported {reaction_type=}")

def check_chemistry(reaction_type, reactants, product, debug=False):

	if reaction_type not in SUPPORTED_CHEMISTRY:

		logger.var("reactants", reactants.ids)
		logger.var("product", product)

		raise UnsupportedChemistryError(f'Unsupported {reaction_type=}')

	assert reactants
	assert product	

	CHEMISTRY = SUPPORTED_CHEMISTRY[reaction_type]
	
	if 'heavy_atoms_diff' in CHEMISTRY:
		check = check_count_diff("heavy_atoms", reaction_type, reactants, product, debug=debug)
		if not check:
			return False

	if 'rings_diff' in CHEMISTRY:
		check = check_count_diff("rings", reaction_type, reactants, product, debug=debug)
		if not check:
			return False

	if 'atomtype' in CHEMISTRY:
		check = check_atomtype_diff(reaction_type, reactants, product, debug=debug)
		if not check:
			return False

	if debug:
		logger.success(f'{reaction_type}: All OK')

	return True

def check_count_diff(check_type, reaction_type, reactants, product, debug=False):

	# get target value
	diff = SUPPORTED_CHEMISTRY[reaction_type][f'{check_type}_diff']

	# get attribute name
	attr = f'num_{check_type}'

	# get values
	reac_count = getattr(reactants, attr)
	prod_count = getattr(product, attr)
	if debug: logger.var(f'#{check_type} reactants', reac_count)
	if debug: logger.var(f'#{check_type} product', prod_count)

	# check against target value
	if reac_count - diff != prod_count:
		if debug: logger.error(f'{reaction_type}: #{check_type} FAIL')
		return False

	else:
		if debug: logger.success(f'{reaction_type}: #{check_type} OK')

	return True

def check_atomtype_diff(reaction_type, reactants, product, debug=False):

	check_type = 'atomtype'

	# get values
	reac = reactants.atomtype_dict
	prod = product.atomtype_dict

	if debug: 
		logger.var('reactants.atomtype_dict', str(reac))
		logger.var('product.atomtype_dict', str(prod))
	
	if 'removed' in SUPPORTED_CHEMISTRY[reaction_type]['atomtype']:
		removal = check_specific_atomtype_diff(reaction_type, prod, reac, removal=True, debug=debug)

		if not removal:
			return False

	if 'added' in SUPPORTED_CHEMISTRY[reaction_type]['atomtype']:
		addition = check_specific_atomtype_diff(reaction_type, prod, reac, removal=False, debug=debug)

		if not addition:
			return False

	if debug: 
		logger.success(f'{reaction_type}: atomtypes OK')
	
	return True

def check_specific_atomtype_diff(reaction_type, prod, reac, removal=False, debug=False):
	
	if removal:
		add_str = 'removed'
	else:
		add_str = 'added'

	add_dict = SUPPORTED_CHEMISTRY[reaction_type]['atomtype'][add_str]

	if not add_dict:
		return True

	if debug: 
		logger.var(add_str, str(add_dict))
	
	for symbol, count in add_dict.items():

		if symbol == 'Ha':
			p_count = halogen_count(prod)
			r_count = halogen_count(reac)

		else:
			p_count = prod[symbol] if symbol in prod else 0
			r_count = reac[symbol] if symbol in reac else 0
		
		if removal and r_count - p_count != count:
			if debug: 
				logger.error(f'{symbol}: {r_count=} - {p_count=} = {r_count - p_count}')
				logger.error(f'{reaction_type}: atomtype removal {symbol} x {count} FAIL')
			return False

		elif not removal and p_count - r_count != count:
			if debug: 
				logger.error(f'{symbol}: {p_count=} - {r_count=} = {p_count - r_count}')
				logger.error(f'{reaction_type}: atomtype addition {symbol} x {count} FAIL')
			return False
	
	return True

def halogen_count(atomtype_dict):
	count = 0
	symbols = ['F', 'Cl', 'Br', 'I']
	for symbol in symbols:
		if symbol in atomtype_dict:
			count += atomtype_dict[symbol]
	return count

class InvalidChemistryError(Exception):
	pass

class UnsupportedChemistryError(Exception):
	pass




import mcol
# from dataclasses import dataclass, field, asdict

import logging
logger = logging.getLogger('HIPPO')

# from typing import TypeVar, Type

# T = TypeVar('T')

# @dataclass
class Quote:
	
	"""Dataclass representing a quote in the database"""

	_db = None

	def __init__(self,
		db,
		id: int,
		compound: int,
		smiles: str,
		supplier: str,
		catalogue: str,
		entry: str,
		amount: float,
		price: float,
		currency: str,
		purity: float,
		lead_time: int,
		date: str | None = None,
		type: str | None = None,
	):

		self._db = db
		self._id = id
		self._compound = compound
		self._smiles = smiles
		self._supplier = supplier
		self._catalogue = catalogue
		self._entry = entry
		self._amount = amount
		self._price = price
		self._currency = currency
		self._purity = purity
		self._lead_time = lead_time
		self._date = date
		self._type = type

		from datetime import datetime
		quote_age = (datetime.today() - datetime.strptime(self.date, '%Y-%m-%d')).days
		if quote_age > 30:
			logger.warning(f'Quote is older than {quote_age} days')
			logger.warning(self)

	### FACTORIES

	@classmethod
	def combination(cls, required_amount, quotes):

		"""

		* Start with biggest pack
		* Estimate by scaling linearly with unit price
		* Use MCule confirmed stock amount

		"""

		biggest_pack = sorted(quotes, key=lambda x: x.amount)[-1]
		unit_price = biggest_pack.price / biggest_pack.amount
		estimated_price = unit_price * required_amount

		self = cls.__new__(cls)
		self.__init__(
			db=biggest_pack.db,
			id=None,
			compound=biggest_pack.compound,
			smiles=biggest_pack.smiles,
			supplier=biggest_pack.supplier,
			catalogue=biggest_pack.catalogue,
			entry=biggest_pack.entry,
			amount=required_amount,
			price=estimated_price,
			currency=biggest_pack.currency,
			purity=biggest_pack.purity,
			lead_time=biggest_pack.lead_time,
			date=biggest_pack.date,
			type=f'estimate from quote={biggest_pack.id}'
		)
		
		return self

	### PROPERTIES

	@property
	def entry_str(self):
		if self.catalogue:
			return f'{self.supplier}:{self.catalogue}:{self.entry}'
		else:
			return f'{self.supplier}:{self.entry}'

	@property
	def db(self):
		return self._db

	@property
	def id(self):
		return self._id

	@property
	def compound(self):
		return self._compound

	@property
	def smiles(self):
		return self._smiles

	@property
	def supplier(self):
		return self._supplier

	@property
	def catalogue(self):
		return self._catalogue

	@property
	def entry(self):
		return self._entry

	@property
	def amount(self):
		return self._amount

	@property
	def price(self):
		return self._price

	@property
	def currency(self):
		return self._currency

	@property
	def purity(self):
		return self._purity

	@property
	def lead_time(self):
		return self._lead_time

	@property
	def date(self):
		return self._date

	@property
	def type(self):
		return self._type

	@property
	def dict(self):
		return dict(
			id=self.id,
			compound=self.compound,
			smiles=self.smiles,
			supplier=self.supplier,
			catalogue=self.catalogue,
			entry=self.entry,
			amount=self.amount,
			price=self.price,
			currency=self.currency,
			purity=self.purity,
			lead_time=self.lead_time,
			date=self.date,
			type=self.type,
		)

	@property
	def currency_symbol(self):
		return {'USD':'$', 'EUR':'€', 'GBP':'£'}[self.currency]
	
	### METHODS

	### DUNDERS

	def __repr__(self):

		if self.purity:
			purity = f' @ {self.purity:.0%}'
		else:
			purity = ''

		if self.supplier == 'Stock':
			return f'{mcol.bold}{mcol.underline}C{self.compound} In Stock: {self.amount:}mg{purity}{mcol.unbold}{mcol.ununderline}'

		if self.type:
			return f'{mcol.bold}{mcol.underline}C{self.compound} {self.entry_str} {self.amount:}mg{purity} = {self.price:} {self.currency} ({self.lead_time} days) {self.smiles} [{self.type}]{mcol.unbold}{mcol.ununderline}'
		else:
			return f'{mcol.bold}{mcol.underline}C{self.compound} {self.entry_str} {self.amount:}mg{purity} = {self.price:} {self.currency} ({self.lead_time} days) {self.smiles}{mcol.unbold}{mcol.ununderline}'




from collections import UserDict
from typing import Mapping

class MetaData(UserDict):

	"""Metadata dictionary linked to a compound or pose in the database"""

	def __init__(self, 
		__dict: Mapping[str, str] | None
	) -> None:

		super().__init__()
		if __dict:
			for key, value in __dict.items():
				super().__setitem__(key, value)

		self._db = None
		self._table: str = None
		self._id: str = None

	def _update_db(self, commit=True):
		self._db.insert_metadata(table=self._table, id=self._id, payload=self.data, commit=commit)

	def __setitem__(self, key: str, item: str) -> None:
		self.data.__setitem__(key, item)
		self._update_db()

	def __delitem__(self, key: str) -> None:
		self.data.__delitem__(key)
		self._update_db()

	def update(self, data, commit=True):
		"""Wrapper for dict.update()"""
		self.data.update(data)
		self._update_db(commit=commit)

	def append(self, key, value):
		"""Create or append to a list-like value with given key"""
		if key not in self:
			self.data[key] = []
		if value not in self.data[key]:
			self.data[key].append(value)
		self._update_db()




from dataclasses import dataclass, field

from .compound import Ingredient

import logging
logger = logging.getLogger('HIPPO')

# @dataclass
class Recipe:
	""" A Recipe stores data corresponding to a specific synthetic recipe involving several products, reactants, intermediates, and reactions."""

	_db = None

	def __init__(self, db, products, reactants, intermediates, reactions):

		from .cset import IngredientSet
		from .rset import ReactionSet
		
		# check typing
		assert isinstance(products, IngredientSet)
		assert isinstance(reactants, IngredientSet)
		assert isinstance(intermediates, IngredientSet)
		assert isinstance(reactions, ReactionSet)
		
		self._products = products
		self._reactants = reactants
		self._intermediates = intermediates
		self._reactions = reactions
		self._db = db

	### FACTORIES


	@classmethod
	def from_reaction(cls, 
		reaction, 
		amount=1, 
		debug: bool = False, 
		pick_cheapest: bool = True,
		permitted_reactions=None,
		quoted_only: bool = False,
		supplier: None | str = None,
	):

		from .reaction import Reaction
		assert isinstance(reaction, Reaction)

		from .cset import IngredientSet
		from .rset import ReactionSet

		if debug: logger.debug(f'Recipe.from_reaction({amount=}, {pick_cheapest=})')
		if debug: logger.debug(f'{reaction.id=}')
		if debug: logger.debug(f'{reaction.product.id=}')
		if debug: logger.debug(f'{amount=}')
		if debug: logger.debug(f'{reaction.reactants.ids=}')

		if permitted_reactions:
			assert reaction in permitted_reactions
			# raise NotImplementedError

		db = reaction.db

		recipe = cls.__new__(cls)
		recipe.__init__(db, 
			products=IngredientSet(db, [reaction.product.as_ingredient(amount=amount)]),
			reactants=IngredientSet(db, []),
			intermediates=IngredientSet(db, []),
			reactions=ReactionSet(db, [reaction.id]),
		)

		recipes = [recipe]

		if quoted_only:
			ok = reaction.check_reactant_availability(supplier=supplier)
			if not ok:
				logger.error(f'Reactants not available for {reaction=}')
				if pick_cheapest:
					return None
				else:
					return []
		
		pairs = reaction.get_reactant_amount_pairs()
		for reactant, reactant_amount in pairs:
			if debug: logger.debug(f'{reactant.id=}, {reactant_amount=}')

			# scale amount
			reactant_amount *= amount
			reactant_amount /= reaction.product_yield

			inner_reactions = reactant.get_reactions(none='quiet', permitted_reactions=permitted_reactions)

			if inner_reactions:

				if debug: 
					if len(inner_reactions) == 1:
						logger.debug(f'Reactant has ONE inner reaction')
					else:
						logger.warning(f'{reactant=} has MULTIPLE inner reactions')

				new_recipes = []

				inner_recipes = []
				for reaction in inner_reactions:
					reaction_recipes = Recipe.from_reaction(reaction=reaction, amount=reactant_amount, debug=debug, pick_cheapest=False, quoted_only=quoted_only)
					inner_recipes += reaction_recipes

				for recipe in recipes:

					for inner_recipe in inner_recipes:

						combined_recipe = recipe.copy()

						combined_recipe.reactants += inner_recipe.reactants
						combined_recipe.intermediates += inner_recipe.intermediates
						combined_recipe.reactions += inner_recipe.reactions
						combined_recipe.intermediates.add(reactant.as_ingredient(reactant_amount))

						new_recipes.append(combined_recipe)
					
				recipes = new_recipes

			else:

				ingredient = reactant.as_ingredient(reactant_amount)
				for recipe in recipes:
					recipe.reactants.add(ingredient)

		if pick_cheapest:
			priced = [r for r in recipes if r.price]
			if not priced:
				logger.error("0 recipes with prices, can't choose cheapest")
				return recipes
			return sorted(priced, key=lambda r: r.price)[0]

		return recipes

	@classmethod
	def from_reactions(cls, 
		reactions, 
		amount=1,
		pick_cheapest: bool = True,
		permitted_reactions=None,
		final_products_only=True,
	):

		from .rset import ReactionSet
		from .cset import IngredientSet
		assert isinstance(reactions, ReactionSet)

		# get all the products
		products = reactions.products

		return products

		if final_products_only:
			ids = reactions.db.execute(f'''
				SELECT DISTINCT compound_id FROM compound
				LEFT JOIN reactant ON compound_id = reactant_compound
				WHERE reactant_compound IS NULL
				AND compound_id IN {products.str_ids}
			''').fetchall()

			return ids

		recipe = Recipe.from_compounds(compounds=products, amount=amount, permitted_reactions=reactions, pick_cheapest=pick_cheapest)

		return recipe

	@classmethod
	def from_compounds(cls,
		compounds, 
		amount: float = 1, 
		debug: bool = False, 
		pick_cheapest: bool = True,
		permitted_reactions=None,
		quoted_only: bool = False,
		supplier: None | str = None,
		solve_combinations: bool = True,
		pick_first: bool = False,
		warn_multiple_solutions: bool = True,
		pick_cheapest_inner_routes: bool = False,
	):

		from tqdm import tqdm

		from .cset import CompoundSet
		assert isinstance(compounds, CompoundSet)

		# if permitted_reactions:
		# 	raise NotImplementedError

		n_comps = len(compounds)

		if not hasattr(amount, '__iter__'):
			amount = [amount] * n_comps

		options = []

		logger.debug(f'{n_comps=}')

		logger.info('Solving individual compound recipes...')
		for comp, a in tqdm(zip(compounds, amount), total=n_comps):

			comp_options = []

			for reaction in comp.reactions:

				if permitted_reactions and reaction not in permitted_reactions:
					continue

				if pick_cheapest_inner_routes:
					sol = Recipe.from_reaction(reaction=reaction, amount=a, pick_cheapest=True, debug=debug, permitted_reactions=permitted_reactions, quoted_only=quoted_only, supplier=supplier)
					if sol:
					# assert isinstance(sol, Recipe)
						comp_options.append(sol)
				else:
					sols = Recipe.from_reaction(reaction=reaction, amount=a, pick_cheapest=False, debug=debug, permitted_reactions=permitted_reactions, quoted_only=quoted_only, supplier=supplier)	
					assert isinstance(sols, list)
					comp_options += sols

			if warn_multiple_solutions and len(comp_options) > 1:
				logger.warning(f'Multiple solutions for compound={comp}')

			if not comp_options:
				logger.error(f'No solutions for compound={comp}')
			else:
				options.append(comp_options)

		assert all(options)

		from itertools import product
		logger.info('Solving recipe combinations...')
		combinations = list(product(*options))

		if not solve_combinations:
			return combinations

		if pick_first:
			combinations = [combinations[0]]
			
		logger.info('Combining recipes...')
		
		solutions = []
		for combo in tqdm(combinations):

			logger.debug(f'Combination of {len(combo)} recipes')

			solution = combo[0]

			for recipe in combo[1:]:
				solution += recipe

			solutions.append(solution)

		if not solutions:
			logger.error('No solutions!')

		if pick_first:
			return solutions[0]

		if pick_cheapest:
			logger.info('Picking cheapest...')
			priced = [r for r in solutions if r.price]
			if not priced:
				logger.error("0 recipes with prices, can't choose cheapest")
				return solutions
			return sorted(priced, key=lambda r: r.price)[0]

		return solutions

	@classmethod
	def from_reactants(cls, reactants, amount=1, debug=False):

		# print(reactants.ids)


		from .cset import IngredientSet

		if isinstance(reactants, IngredientSet):
			reactant_ids = reactants.compound_ids
		else:
			reactant_ids = reactants.ids

		db = reactants.db

		compound_ids = reactant_ids

		possible_reactions = []

		# recursively search for possible reactions
		for i in range(300):
			
			reaction_ids = db.get_possible_reaction_ids(compound_ids=compound_ids)

			if not reaction_ids:
				break

			possible_reactions += reaction_ids

			if debug:
				logger.var('reaction_ids', reaction_ids)

			product_ids = db.get_possible_reaction_product_ids(reaction_ids=reaction_ids)

			if debug:
				logger.var('product_ids', product_ids)

			compound_ids = product_ids

		else:
			raise NotImplementedError('Maximum recursion depth exceeded')

		if debug:
			logger.var('all possible reactions', possible_reactions)

		from .rset import ReactionSet

		rset = ReactionSet(db, possible_reactions)

		return cls.from_reactions(rset, amount=amount, permitted_reactions=rset, pick_cheapest=False)

	@classmethod
	def from_ingredients(cls, db, ingredients):
		raise NotImplementedError
		self = cls.__new__(cls)
		self.__init__(...)
		return self

	### PROPERTIES

	@property
	def db(self):
		return self._db

	@property
	def products(self):
		return self._products

	@products.setter
	def products(self, a):
		self._products = a

	@property
	def reactants(self):
		return self._reactants

	@reactants.setter
	def reactants(self, a):
		self._reactants = a

	@property
	def intermediates(self):
		return self._intermediates

	@intermediates.setter
	def intermediates(self, a):
		self._intermediates = a

	@property
	def reactions(self):
		return self._reactions

	@reactions.setter
	def reactions(self, a):
		self._reactions = a

	# @property
	# def quotes(self):
		# return self.get_quotes()

	@property
	def price(self):
		# total = 0
		# quotes = self.quotes
		# if not quotes:
		# 	return None
		# assert len((currencies := set([q.currency for q in quotes]))) == 1, 'Multiple currencies'
		# return sum([q.price for q in quotes]), list(currencies)[0]
		return self.reactants.price

	# @property
	# def lead_time(self):
	# 	total = 0
	# 	quotes = self.quotes
	# 	if not quotes:
	# 		return None
	# 	return max([q.lead_time for q in quotes])
	
	### METHODS

	def draw(self, color_mapper = None, node_size=300, graph_only=False):

		"""draw graph of the reaction network"""

		import networkx as nx

		color_mapper = color_mapper or {} 
		colors = {}
		sizes = {}

		graph = nx.DiGraph()

		for reaction in self.reactions:
			for reactant in reaction.reactants:
				key = str(reactant)
				ingredient = self.get_ingredient(id=reactant.id)

				graph.add_node(key, id=reactant.id, smiles=reactant.smiles, amount=ingredient.amount, price=ingredient.price, lead_time=ingredient.lead_time)

				if not graph_only:
					sizes[key] = self.get_ingredient(id=reactant.id).amount
					if key in color_mapper:
						colors[key] = color_mapper[key]
					else:
						colors[key] = (0.7,0.7,0.7)

		for product in self.products:
			key = str(product.compound)			
			ingredient = self.get_ingredient(id=product.id)

			graph.add_node(key, id=product.id, smiles=product.smiles, amount=ingredient.amount, price=ingredient.price, lead_time=ingredient.lead_time)

			if not graph_only:
				sizes[key] = product.amount
				if key in color_mapper:
					colors[key] = color_mapper[key]
				else:
					colors[key] = (0.7,0.7,0.7)

		for reaction in self.reactions:
			for reactant in reaction.reactants:
				graph.add_edge(str(reactant), str(reaction.product), id=reaction.id, type=reaction.type, product_yield=reaction.product_yield)

		# rescale sizes
		if not graph_only:
			s_min = min(sizes.values())
			sizes = [s/s_min*node_size for s in sizes.values()]

		if graph_only:
			return graph
		else:
			import matplotlib as plt
			# return nx.draw(graph, pos, with_labels=True, font_weight='bold')
			# pos = nx.spring_layout(graph, iterations=200, k=30)
			pos = nx.spring_layout(graph)
			return nx.draw(graph, pos=pos, with_labels=True, font_weight='bold', node_color=list(colors.values()), node_size=sizes)

	def sankey(self, title=None):
		"""draw a plotly Sankey diagram"""

		graph = self.draw(graph_only=True)

		import plotly.graph_objects as go

		nodes = {}

		for edge in graph.edges:

			c = edge[0]
			if c not in nodes:
				nodes[c] = len(nodes)

			c = edge[1]
			if c not in nodes:
				nodes[c] = len(nodes)

		source = [nodes[a] for a,b in graph.edges]
		target = [nodes[b] for a,b in graph.edges]
		value = [1 for l in graph.edges]

		# print(graph.nodes)
		
		labels = list(nodes.keys())
		
		# compound_ids = [n.id for n in nodes]
		# smiles = [n.smiles for n in nodes]
		# customdata = [(n.id, n.smiles) for n in ]

		hoverkeys = None

		customdata = []
		for key in nodes.keys():
			n = graph.nodes[key]
			
			if not hoverkeys:
				hoverkeys = list(n.keys())
			
			if not n:
				logger.error(f'problem w/ node {key=}')
				compound_id = int(key[1:])
				customdata.append((compound_id, None))

			else:
				# customdata.append((n['id'], n['smiles']))
				d = tuple(v if v is not None else 'N/A' for v in n.values())
				customdata.append(d)
				# id=product.id, smiles=product.smiles, amount=ingredient.amount, price=ingredient.price, lead_time=ingredient.lead_time

		hoverkeys_edges = None
		
		# edgedata = [graph.edges[a,b]["reaction_id"] for a,b in graph.edges]

		customdata_edges = []

		for s,t in graph.edges.keys():
			edge = graph.edges[s,t]
			
			if not hoverkeys_edges:
				hoverkeys_edges = list(edge.keys())
			
			if not n:
				logger.error(f'problem w/ edge {s=} {t=}')
				customdata_edges.append((None, None, None))

			else:
				d = tuple(v if v is not None else 'N/A' for v in edge.values())
				customdata_edges.append(d)

		hoverlines = []
		for i,key in enumerate(hoverkeys):
			hoverlines.append(f'{key}=%''{'f'customdata[{i}]''}')
		hovertemplate = 'Compound '+'<br>'.join(hoverlines)+'<extra></extra>'

		hoverlines_edges = []
		for i,key in enumerate(hoverkeys_edges):
			hoverlines_edges.append(f'{key}=%''{'f'customdata[{i}]''}')
		hovertemplate_edges = 'Reaction '+'<br>'.join(hoverlines_edges)+'<extra></extra>'

		# print(hovertemplate)

		# compound_ids = [int(s[1:]) for s in labels]

		# from .cset import CompoundSet
		# smiles = CompoundSet(self.db, compound_ids).smiles

		# print(compound_ids)

		fig = go.Figure(data=[go.Sankey(
				node = dict(
				# pad = 15,
				# thickness = 20,
				# line = dict(color = "black", width = 0.5),
				label = labels,
				# color = "blue"
				customdata=customdata,
				# customdata = ["Long name A1", "Long name A2", "Long name B1", "Long name B2",
					# "Long name C1", "Long name C2"],
				# hovertemplate='Compound %{label}<br><br>smiles=%{customdata}<extra></extra>',
				hovertemplate=hovertemplate,
			),
				link = dict(
				customdata=customdata_edges,
				hovertemplate=hovertemplate_edges,
				source = source,
				target = target,
				value = value,
		))])

		if not title:
			# title = f"Recipe<br><sup>price={self.price}, lead-time={self.lead_time}</sup>"
			title = f"Recipe<br><sup>price={self.price}</sup>"

		fig.update_layout(title=title)

# link = dict(
#       source = [0, 1, 0, 2, 3, 3], # indices correspond to labels, eg A1, A2, A2, B1, ...
#       target = [2, 3, 3, 4, 4, 5],
#       value = [8, 4, 2, 8, 4, 2],
#       customdata = ["q","r","s","t","u","v"],
#       hovertemplate='Link from node %{source.customdata}<br />'+
#         'to node%{target.customdata}<br />has value %{value}'+
#         '<br />and data %{customdata}<extra></extra>',
#   )

		return fig

	def summary(self):
		"""Print a summary of this recipe"""

		import mcol
		
		logger.header('Recipe')

		price = self.price
		if price:
			logger.var('\nprice', price[0], dict(unit=price[1]))
			# logger.var('lead-time', self.lead_time, dict(unit='working days'))

		logger.var('\n#products', len(self.products))
		for product in self.products:
			logger.var(str(product.compound), f'{product.amount:.2f}', dict(unit='mg'))

		logger.var('\n#intermediates', len(self.intermediates))
		for intermediate in self.intermediates:
			logger.var(str(intermediate.compound), f'{intermediate.amount:.2f}', dict(unit='mg'))

		logger.var('\n#reactants', len(self.reactants))
		for reactant in self.reactants:
			logger.var(str(reactant.compound), f'{reactant.amount:.2f}', dict(unit='mg'))
			# logger.out(f'{mcol.varName}{reactant}{mcol.clear} = {reactant.amount:.2f} {mcol.varType}mg{mcol.clear}, {self.get_reactant_reactions(reactant)}')

		logger.var('\n#reactions', len(self.reactions))
		for reaction in self.reactions:
			logger.var(str(reaction), reaction.reaction_str, dict(unit=reaction.type))

	def get_ingredient(self, id):
		"""Get an ingredient by its compound ID"""
		matches = [r for r in self.reactants + self.products + self.intermediates if r.id == id]
		assert len(matches) == 1
		return matches[0]

	def add_to_all_reactants(self, amount=20):
		self.reactants.df['amount'] += amount

	# def get_quotes(self, df=False):

	# 	quotes = [quote for i in self.reactants if (quote := i.quote)]

	# 	if df:
	# 		from pandas import DataFrame
	# 		return DataFrame([q.dict for q in quotes])

	# 	return quotes

	def write_json(self, file, indent=4):

		"""Serialise this recipe object and write it to disk

		Store
		=====

		- Path to database
		- Timestamp
		- Reactants (& their quotes, amounts)
		- Intermediates (& their quotes)
		- Products (& their poses/scores/fingerprints)
		- Reactions
		- Total Price
		- Lead time

		"""

		import json
		from datetime import datetime

		data = {}

		# Database
		data['database'] = str(self.db.path.resolve())
		data['timestamp'] = str(datetime.now())

		# Recipe properties
		data['price'] = self.price
		# data['lead_time'] = self.lead_time

		# IngredientSets
		data['reactants'] = self.reactants.df.to_dict(orient='records')
		data['intermediates'] = self.intermediates.df.to_dict(orient='records')
		data['products'] = self.products.df.to_dict(orient='records')
		
		# ReactionSet
		data['reaction_ids'] = self.reactions.ids

		# return data

		logger.writing(file)
		json.dump(data, open(file, 'wt'), indent=indent)

	def copy(self):
		return Recipe(self.db, self.products.copy(), self.reactants.copy(), self.intermediates.copy(), self.reactions.copy())

	# def get_reactant_reactions(self, reactant):
	# 	"""Get reactions that a reactant is involved in"""
	# 	return [r for r in self.reactions if reactant in r.reactants]

	# def get_reactant_reaction_string(self, reactant):
	# 	reactions = self.get_reactant_reactions(reactant)
	# 	return ' '.join([str(r) for r in reactions])

	### DUNDERS

	def __repr__(self):
		return f'Recipe({self.reactants} --> {self.intermediates} --> {self.products} via {self.reactions})'
		# return f'Recipe()'

	def __add__(self, other):
		result = self.copy()
		result.reactants += other.reactants
		result.intermediates += other.intermediates
		result.reactions += other.reactions
		result.products += other.products
		return result





# from .tools import df_row_to_dict

from .compound import Compound, Ingredient
from .db import Database

from .recipe import Recipe

from numpy import int64

import mcol

import os

import logging
logger = logging.getLogger('HIPPO')

class CompoundTable:
	"""Object representing the 'compound' table in the :class:`.Database`."""

	def __init__(self, 
		db: Database, 
		table: str = 'compound',
	) -> None:
		
		self._db = db
		self._table = table

	### FACTORIES

	### PROPERTIES

	@property
	def db(self) -> Database:
		"""Returns the associated :class:`.database`"""
		return self._db
	
	@property
	def table(self) -> str:
		return self._table

	@property
	def names(self) -> list:
		"""Returns the names of child compounds"""
		result = self.db.select(table=self.table, query='compound_name', multiple=True)
		return [q for q, in result]
	
	@property
	def ids(self):
		"""Returns the IDs of child compounds"""
		result = self.db.select(table=self.table, query='compound_id', multiple=True)
		return [q for q, in result]
	
	@property
	def tags(self):
		"""Returns the set of unique tags present in this compound set"""
		values = self.db.select_where(table='tag', query='DISTINCT tag_name', key='tag_compound IS NOT NULL', multiple=True)
		return set(v for v, in values)

	@property
	def reactants(self):
		"""Returns a CompoundSet of all compounds that are used as a reactants"""
		# ids = self.db.select(table='reactant', query='DISTINCT reactant_compound', multiple=True)
		ids = self.db.execute('SELECT reactant_compound FROM reactant LEFT JOIN reaction ON reactant.reactant_compound = reaction.reaction_product WHERE reaction.reaction_product IS NULL').fetchall()
		ids = [q for q, in ids]
		from .cset import CompoundSet
		return CompoundSet(self.db, ids)

	@property
	def products(self):
		"""Returns a CompoundSet of all compounds that are a product of a reaction but not a reactant"""
		ids = self.db.execute('SELECT reaction_product FROM reaction LEFT JOIN reactant ON reaction.reaction_product = reactant.reactant_compound WHERE reactant.reactant_compound IS NULL').fetchall()
		ids = [q for q, in ids]
		from .cset import CompoundSet
		return CompoundSet(self.db, ids)

	@property
	def intermediates(self):
		"""Returns a CompoundSet of all compounds that are products and reactants"""
		ids = self.db.execute('SELECT DISTINCT reaction_product FROM reaction INNER JOIN reactant ON reaction.reaction_product = reactant.reactant_compound').fetchall()
		ids = [q for q, in ids]
		from .cset import CompoundSet
		return CompoundSet(self.db, ids)

	@property
	def num_reactants(self):
		"""Returns the number of reactants (see HIPPO.reactants)"""
		return len(self.reactants)

	@property
	def num_intermediates(self):
		"""Returns the number of intermediates (see HIPPO.intermediates)"""
		return len(self.intermediates)
	
	@property
	def num_products(self):
		"""Returns the number of products (see HIPPO.products)"""
		return len(self.products)

	@property
	def elabs(self):
		"""Returns a CompoundSet of all compounds that are a an elaboration of an existing base"""
		ids = self.db.select_where(query='compound_id', table='compound', key='compound_base IS NOT NULL', multiple=True)
		ids = [q for q, in ids]
		from .cset import CompoundSet
		return CompoundSet(self.db, ids)

	@property
	def bases(self):
		"""Returns a CompoundSet of all compounds that are the basis for a set of elaborations"""
		ids = self.db.select_where(query='DISTINCT compound_base', table='compound', key='compound_base IS NOT NULL', multiple=True, none='quiet')
		ids = [q for q, in ids]
		from .cset import CompoundSet
		return CompoundSet(self.db, ids)

	@property
	def num_elabs(self):
		"""Returns the number of compounds that are a an elaboration of an existing base"""
		return len(self.elabs)

	@property
	def num_bases(self):
		"""Returns the number of compounds that are the basis for a set of elaborations"""
		return len(self.bases)

	### METHODS

	def get_by_tag(self,tag):
		"""Get all child compounds with a certain tag"""
		values = self.db.select_where(query='tag_compound', table='tag', key='name', value=tag, multiple=True)
		ids = [v for v, in values if v]
		return self[ids]

	def get_by_metadata(self, key: str, value: str | None = None):
		"""Get all child compounds with by their metadata. If no value is passed, then simply containing the key in the metadata dictionary is sufficient"""
		results = self.db.select(query='compound_id, compound_metadata', table='compound', multiple=True)
		if value is None:
			ids = [i for i,d in results if d and f'"{key}":' in d]
		else:
			if isinstance(value, str):
				value = f'"{value}"'
			ids = [i for i,d in results if d and f'"{key}": {value}' in d]
		return self[ids]		

	def get_by_base(self,base):

		if not isinstance(base, int):
			assert base._table == 'compound'
			base = base.id

		values = self.db.select_where(query='compound_id', table='compound', key='base', value=base, multiple=True)
		ids = [v for v, in values if v]
		return self[ids]
	
	def summary(self):
		"""Print a summary of this compound set"""
		logger.header('CompoundTable()')
		logger.var('#compounds', len(self))
		# logger.var('#poses', self.num_poses)
		logger.var('tags', self.tags)
		logger.var('#bases', self.num_bases)
		logger.var('#elabs', self.num_elabs)
		logger.var('#reactants', self.num_reactants)
		logger.var('#intermediates', self.num_intermediates)
		logger.var('#products', self.num_products)

	def draw(self):
		return self[self.ids].draw()

	def interactive(self):
		return self[self.ids].interactive()

	### DUNDERS

	def __call__(self, *, tag=None, base=None):
		if tag:
			return self.get_by_tag(tag)
		elif base:
			return self.get_by_base(base)
		else:
			raise NotImplementedError

	def __getitem__(self, key) -> Compound:
		
		match key:

			# case int():
			case key if isinstance(key, int) or isinstance(key, int64):

				if key == 0:
					return self.__getitem__(key=1)

				if key < 0:
					key = len(self) + 1 + key
					return self.__getitem__(key=key)

				else:
					return self.db.get_compound(table=self.table, id=key)

			case str():
				comp = self.db.get_compound(table=self.table, inchikey=key)
				if not comp:
					comp = self.db.get_compound(table=self.table, alias=key)
				return comp

			case key if isinstance(key, list) or isinstance(key, tuple) or isinstance(key, set):

				indices = []
				for i in key:
					if isinstance(i,int):
						index = i
					elif isinstance(i,str):
						index = self.db.get_compound_id(name=i)
					else:
						raise NotImplementedError

					assert index
					indices.append(index)

				return CompoundSet(self.db, indices)

			case slice():
				ids = self.db.slice_ids(table=self.table, start=key.start, stop=key.stop, step=key.step)
				return self[ids]

			case _:
				logger.error(f'Unsupported type for CompoundTable.__getitem__(): {key=} {type(key)}')

		return None

	def __repr__(self) -> str:
		return f'{mcol.bold}{mcol.underline}''{'f'C x {len(self)}''}'f'{mcol.unbold}{mcol.ununderline}'

	def __len__(self) -> int:
		return self.db.count(self.table)

	def __iter__(self):
		return iter(self[i+1] for i in range(len(self)))


class CompoundSet:
	"""Object representing a subset of the 'compound' table in the :class:`.Database`."""

	_table = 'compound'
	
	def __init__(self,
		db: Database,
		indices: list = None,
		sort: bool = True,
	):

		self._db = db

		indices = indices or []

		if not isinstance(indices, list):
			indices = list(indices)

		# print(indices)
		# print([type(i) for i in indices])

		indices = [int(i) for i in indices]

		# assert all(isinstance(i, int) for i in indices), indices

		if sort:
			self._indices = sorted(list(set(indices)))
		else:
			self._indices = list(set(indices))

	### PROPERTIES

	@property
	def db(self):
		return self._db

	@property
	def table(self):
		return self._table

	@property
	def indices(self):
		"""Returns the ids of compounds in this set"""
		return self._indices

	@property
	def ids(self):
		"""Returns the ids of compounds in this set"""
		return self.indices

	@property
	def names(self):
		"""Returns the aliases of compounds in this set"""
		result = self.db.select_where(query='compound_alias', table='compound', key=f'compound_id in {self.str_ids}', multiple=True)
		return [q for q, in result]

	@property
	def smiles(self) -> list:
		"""Returns the smiles of child compounds"""
		result = self.db.select_where(query='compound_smiles', table='compound', key=f'compound_id in {self.str_ids}', multiple=True)
		return [q for q, in result]
	
	@property
	def inchikeys(self):
		"""Returns the inchikeys of compounds in this set"""
		result = self.db.select_where(query='compound_inchikey', table='compound', key=f'compound_id in {self.str_ids}', multiple=True)
		return [q for q, in result]

	@property
	def tags(self):
		"""Returns the set of unique tags present in this compound set"""
		values = self.db.select_where(table='tag', query='DISTINCT tag_name', key=f'tag_compound in {self.str_ids}', multiple=True)
		if not values:
			return set()
		return set(v for v, in values)

	@property
	def num_poses(self):
		"""Count the poses associated to this set of compounds"""
		from .pset import PoseSet
		return self.db.count_where(table='pose', key=f'pose_compound in {self.str_ids}')

	@property
	def poses(self):
		"""Get the poses associated to this set of compounds"""
		from .pset import PoseSet
		ids = self.db.select_where(query='pose_id', table='pose', key=f'pose_compound in {self.str_ids}', multiple=True)
		ids = [v for v, in ids]
		return PoseSet(self.db, ids)

	@property
	def best_placed_poses(self):
		"""Get the best placed pose for each compound in this set"""
		from .pset import PoseSet
		query = self.db.select_where(table='pose', query='pose_id, MIN(pose_distance_score)', key=f'pose_compound in {self.str_ids} GROUP BY pose_compound', multiple=True)
		ids = [i for i,s in query]
		return PoseSet(self.db, ids)

	@property
	def str_ids(self):
		return str(tuple(self.ids)).replace(',)',')')

	@property
	def num_heavy_atoms(self):
		return sum([c.num_heavy_atoms for c in self])

	@property
	def num_rings(self):
		return sum([c.num_rings for c in self])

	@property
	def formula(self):
		from .tools import atomtype_dict_to_formula
		return atomtype_dict_to_formula(self.atomtype_dict)

	@property
	def atomtype_dict(self):
		from .tools import formula_to_atomtype_dict, combine_atomtype_dicts
		atomtype_dicts = [c.atomtype_dict for c in self]
		return combine_atomtype_dicts(atomtype_dicts)

	### FILTERING

	def get_by_tag(self,tag):
		"""Get all child compounds with a certain tag"""
		values = self.db.select_where(query='tag_compound', table='tag', key='name', value=tag, multiple=True)
		ids = [v for v, in values if v and v in self.ids]
		return CompoundSet(self.db, ids)

	def get_by_metadata(self, key: str, value: str | None = None):
		"""Get all child compounds with by their metadata. If no value is passed, then simply containing the key in the metadata dictionary is sufficient"""
		results = self.db.select(query='compound_id, compound_metadata', table='compound', multiple=True)
		if value is None:
			ids = [i for i,d in results if d and f'"{key}":' in d and i in self.ids]
		else:
			if isinstance(value, str):
				value = f'"{value}"'
			ids = [i for i,d in results if d and f'"{key}": {value}' in d and i in self.ids]
		return CompoundSet(self.db, ids)		

	def get_all_possible_reactants(self, debug=False):
		"""Recursively searches for all the reactants that could possible be needed to synthesise these compounds."""
		all_reactants, all_recipes = self.db.get_unsolved_reaction_tree(product_ids=self.ids, debug=debug) 
		return all_reactants

	### OUTPUT

	def draw(self):
		"""Draw a grid of all contained molecules"""
		from molparse.rdkit import draw_grid

		data = [(str(c), c.mol) for c in self]

		mols = [d[1] for d in data]
		labels = [d[0] for d in data]

		display(draw_grid(mols, labels=labels))

	def grid(self):
		self.draw()

	def summary(self):
		"""Print a summary of this compound set"""
		logger.header('CompoundSet()')
		logger.var('#compounds', len(self))
		logger.var('#poses', self.num_poses)
		logger.var('tags', self.tags)

	def interactive(self):
		"""Creates a ipywidget to interactively navigate this PoseSet."""

		from ipywidgets import interactive, BoundedIntText, Checkbox, interactive_output, HBox, GridBox, Layout, VBox
		from IPython.display import display
		from pprint import pprint

		a = BoundedIntText(
				value=0,
				min=0,
				max=len(self)-1,
				step=1,
				description=f'Comp (/{len(self)}):',
				disabled=False,
			)

		b = Checkbox(description='Name', value=True)
		c = Checkbox(description='Summary', value=False)
		d = Checkbox(description='2D', value=True)
		e = Checkbox(description='poses', value=False)
		f = Checkbox(description='reactions', value=False)
		g = Checkbox(description='Metadata', value=False)

		ui = GridBox([b, c, d, e, f, g], layout=Layout(grid_template_columns="repeat(6, 100px)"))
		ui = VBox([a, ui])
		
		def widget(i, name=True, summary=True, draw=True, poses=True, reactions=True, metadata=True):
			comp = self[i]
			
			if name and not summary:
				print(repr(comp))

			if summary: 
				comp.summary(metadata=False, draw=False)
			
			if draw: 
				comp.draw()
			
			if poses and comp.poses: 
				comp.poses.draw()

			if reactions and (reactions := comp.reactions):
				for r in reactions:
					print(repr(r))
					r.draw()
			
			if metadata:
				logger.title('Metadata:')
				pprint(comp.metadata)

		out = interactive_output(widget, {'i': a, 'name': b, 'summary': c, 'draw':d, 'poses':e, 'reactions':f, 'metadata':g})

		display(ui, out)

	### OTHER METHODS

	def add(self, compound):
		"""Add a compound to this set"""
		if isinstance(compound, Compound):
			compound = compound.id

		if compound not in self.ids:
			from bisect import insort
			insort(self.ids, compound)
	
	def get_recipes(self, 
		amount: float = 1, 
		debug=False,
		pick_cheapest: bool = False,
		permitted_reactions=None,
		quoted_only: bool = False,
		supplier: None | str = None,
		**kwargs,
	):
		"""Generate the :class:`.Recipe` to make these compounds."""
		from .recipe import Recipe
		return Recipe.from_compounds(self, amount=amount, debug=debug, pick_cheapest=pick_cheapest, permitted_reactions=permitted_reactions, quoted_only=quoted_only, supplier=supplier, **kwargs)

	def copy(self):
		return CompoundSet(self.db, self.ids)
	
	def get_df(self, mol=False, reactions=False, metadata=False, count_by_target=False, poses=False, **kwargs):

		from tqdm import tqdm
		from pandas import DataFrame

		data = []

		for comp in tqdm(self):
			d = comp.get_dict(mol=mol, reactions=reactions, metadata=metadata, count_by_target=count_by_target, poses=poses, **kwargs)
			data.append(d)

		return DataFrame(data)

	def get_quoted(self, *, supplier='any'):
		"""Get all member compounds that have a quote from given supplier"""

		if supplier != 'any':
			key = f'quote_compound IN {self.str_ids} AND quote_supplier = "{supplier}"'
		else:
			key = f'quote_compound IN {self.str_ids}'

		ids = self.db.select_where(
			table='quote', 
			query='DISTINCT quote_compound', 
			key=key, 
			multiple=True,
		)
		
		ids = [i for i, in ids]
		return CompoundSet(self.db, ids)

	def get_unquoted(self, *, supplier='any'):
		"""Get all member compounds that do not have a quote from given supplier"""
		quoted = self.get_quoted(supplier=supplier)
		return self - quoted

	def write_smiles_csv(self, file):
		from pandas import DataFrame
		smiles = self.smiles
		df = DataFrame(dict(smiles=smiles))
		df.to_csv(file)

	### DUNDERS

	# def __repr__(self) -> str:
		# return f'{mcol.bold}{mcol.underline}''{'f'C x {len(self)}''}'f'{mcol.unbold}{mcol.ununderline}'

	def __len__(self) -> int:
		return len(self.indices)

	def __iter__(self):
		return iter(self.db.get_compound(table=self.table, id=i) for i in self.indices)

	def __getitem__(self, key) -> Compound:
		match key:
			case int():
				index = self.indices[key]
				return self.db.get_compound(table=self.table, id=index)
			
			case slice():
				indices = self.indices[key]
				return	CompoundSet(self.db, indices)

			case _:
				raise NotImplementedError	

	def __sub__(self, other):

		match other:
		
			case CompoundSet():
				ids = set(self.ids) - set(other.ids)
				return CompoundSet(self.db, ids)

			case IngredientSet():
				logger.warning('Subtracting IngredientSet from CompoundSet. Ignoring quote/amount data')
				ids = set(self.ids) - set([int(i) for i in other.compound_ids])
				print(self.ids)
				print(ids)
				print([int(i) for i in other.compound_ids])
				return CompoundSet(self.db, ids)
				
			case _:
				raise NotImplementedError

	def __add__(self, other):

		match other:

			case Compound():
				return self.add(other)

			case int():
				return self.add(other)
		
			case CompoundSet():
				ids = set(self.ids) | set(other.ids)
				return CompoundSet(self.db, ids)

			case IngredientSet():
				ids = set(self.ids) | set(other.compound_ids)
				return CompoundSet(self.db, ids)
				
			case _:
				raise NotImplementedError

	def __xor__(self, other):

		match other:
		
			case CompoundSet():
				ids = set(self.ids) ^ set(other.ids)
				return CompoundSet(self.db, ids)
				
			case _:
				raise NotImplementedError

	def __repr__(self) -> str:
		return f'{mcol.bold}{mcol.underline}''{'f'C x {len(self)}''}'f'{mcol.unbold}{mcol.ununderline}'


class IngredientSet:

	"""Refactor as a wrapper to a dataframe???"""

	_columns = [
		'compound_id',
		'amount',
		'quote_id',
		'supplier',
		'max_lead_time',
	]
	
	def __init__(self, db, ingredients=None):

		from pandas import DataFrame

		ingredients = ingredients or []

		self._db = db
		
		self._data = DataFrame(columns=self._columns)
		
		for ingredient in ingredients:
			self.add(ingredient)

	@classmethod
	def from_ingredient_df(cls, db, df):
		self = cls.__new__(cls)
		self._db = db
		self._data = df.copy()
		return self

	### PROPERTIES

	@property
	def df(self):
		return self._data

	@property
	def db(self):
		return self._db

	@property
	def price_df(self):
		df = self.df.copy()
		df['price'] = [i.price for i in self]
		return df

	@property
	def price(self):
		pairs = { i:q for i,q in enumerate(self.df['quote_id'])}

		quote_ids = [q for q in pairs.values() if q is not None]

		if quote_ids:

			quote_id_str = str(tuple(quote_ids)).replace(',)', ')')

			result = self.db.select_where(query='quote_price, quote_currency', table='quote', key=f'quote_id in {quote_id_str}', multiple=True)

			prices = [a for a,b in result]
			quoted = sum(prices)
			
			currencies = list(set([b for a,b in result]))
			assert len(currencies) == 1

		else:

			quoted = 0
			currencies = ('USD')


		unquoted = [i for i,q in pairs.items() if q is None]
		unquoted = [self[i].price for i in unquoted]


		# logger.warning(f"{len([p for p in unquoted if p is None])} ingredients don't have a price")

		return (quoted + sum([p for p in unquoted if p]), currencies)

	@property
	def smiles(self):
		compound_ids = list(self.df['compound_id'])
		result = self.db.select_where(query='compound_smiles', table='compound', key=f'compound_id in {tuple(compound_ids)}', multiple=True)
		return [q for q, in result]

	@property
	def compound_ids(self):
		return list(self.df['compound_id'].values)

	@property
	def ids(self):
		return self.compound_ids

	@property
	def str_compound_ids(self):
		return str(tuple(self.df['compound_id'].values)).replace(',)',')')

	@property
	def compounds(self):
		return CompoundSet(self.db, self.compound_ids)

	### METHODS

	def add(self, ingredient=None, *, compound_id=None, amount=None, quote_id=None, supplier=None, max_lead_time=None):

		from pandas import DataFrame, concat

		if ingredient:
			assert ingredient._table == 'ingredient'
			compound_id = ingredient.compound_id
			amount = ingredient.amount

			if ingredient.quote and not ingredient.quote_id:
				logger.warning(f'Losing quote! {ingredient.quote=}')

			quote_id = ingredient.quote_id
			
			supplier = ingredient.supplier
			max_lead_time = ingredient.max_lead_time
			
		else:
			assert compound_id
			assert amount

		if self._data.empty:
			addition = DataFrame([dict(compound_id=compound_id, amount=amount, quote_id=quote_id, supplier=supplier, max_lead_time=max_lead_time)])
			self._data = addition

		else:

			if compound_id in self._data['compound_id'].values:
				index = self._data.index[self._data['compound_id'] == compound_id].tolist()[0]
				self._data.loc[index, 'amount'] += amount
				self._data.loc[index, 'quote_id'] = None

			else:
				from numpy import nan
				addition = DataFrame([dict(compound_id=compound_id, amount=amount, quote_id=quote_id, supplier=supplier, max_lead_time=max_lead_time)])
				self._data = concat([self._data, addition], ignore_index=True)
				self._data = self._data.replace({nan: None})

	def get_ingredient(self, series):
		return Ingredient(
			db=self._db,
			compound=series['compound_id'],
			amount=series['amount'],
			quote=series['quote_id'],
			supplier=series['supplier'],
			max_lead_time=series['max_lead_time'],
		)

	def copy(self):
		return IngredientSet.from_ingredient_df(self.db, self.df)

	def draw(self):
		self.compounds.draw()

	def set_amounts(self, amount):

		self.df['amount'] = amount

		# if amounts are modified the quotes should be cleared
		self.df['quote_id'] = None

		assert all(self.df['supplier'].isna()) and all(self.df['max_lead_time'].isna())

		# update quotes
		pairs = self.db.execute(f'''
			WITH matching_quotes AS (
				SELECT quote_id, quote_compound, MIN(quote_price) FROM quote
				WHERE quote_compound IN {self.str_compound_ids}
				AND quote_amount >= {amount}
				GROUP BY quote_compound
			)
			SELECT compound_id, quote_id FROM compound
			LEFT JOIN matching_quotes ON quote_compound = compound_id
			WHERE compound_id IN {self.str_compound_ids}
		''').fetchall()

		for compound_id, quote_id in pairs:
			match = self.df.index[self.df['compound_id'] == compound_id][0]
			self.df.loc[match, 'quote_id'] = quote_id

	### DUNDERS

	def __len__(self):
		return len(self._data)

	def __repr__(self):
		return f'{mcol.bold}{mcol.underline}''{'f'I x {len(self)}''}'f'{mcol.unbold}{mcol.ununderline}'

	def __add__(self, other):
		for i,row in other._data.iterrows():
			self.add(
				compound_id=row.compound_id,
				amount=row.amount,
				quote_id=row.quote_id,
				supplier=row.supplier,
				max_lead_time=row.max_lead_time,
			)
		return self

	def __getitem__(self, key):
		match key:
			case int():
				series = self.df.loc[key]
				return self.get_ingredient(series)
		
			case _:
				raise NotImplementedError

	def __iter__(self):
		return iter(self.get_ingredient(s) for i,s in self.df.iterrows())

	def __call__(self, *, compound_id):

		if compound_id:

			# get the ingredient with the matching compound ID
			matches = self.df[self.df['compound_id'] == compound_id]

			if len(matches) != 1:

				logger.warning(f'Multiple ingredients in set with {compound_id=}')
				# print(matches)

				return IngredientSet(self.db, [self.get_ingredient(s) for i,s in matches.iterrows()])

			return self.get_ingredient(matches.iloc[0])

		else:
			raise NotImplementedError




import mcol

from rdkit import Chem

from .pose import Pose
# from .pset import PoseSet
from .tags import TagSet
# from .rset import ReactionSet
from .target import Target
from .quote import Quote

import logging
logger = logging.getLogger('HIPPO')


class Compound:

	"""A :class:`.Compound` represents a ligand/small molecule with stereochemistry removed and no atomic coordinates. I.e. it represents the chemical structure. It's name is always an InChiKey. If a compound is an elaboration it can have a :meth:`.Compound.base` property which is another :class:`.Compound`. :class:`.Compound` objects are target-agnostic and can be linked to any number of catalogue entries (:class:`.Quote`) or synthetic pathways (:class:`.Reaction`).

	:class:`.Compound` objects should not be created directly. Instead use :meth:`.HIPPO.register_compound` or :meth:`.HIPPO.compounds`
	"""

	def __init__(self,
			animal,
			db,
			id: int,
			inchikey: str,
			alias: str,
			smiles: str,
			base: int,
			mol: Chem.Mol | bytes | None = None,
			metadata: dict | None = None,
	):
		
		# from compound table
		self._id = id
		self._inchikey = inchikey
		self._alias = alias
		self._smiles = smiles
		self._animal = animal
		self._base_id = base
		self._base = None
		self._alias = alias
		self._tags = None
		self._table = 'compound'
		self._metadata = metadata
		
		# computed properties
		self._num_heavy_atoms = None
		self._num_rings = None
		self._formula = None
		self._molecular_weight = None
		
		if isinstance(mol, bytes):
			mol = Chem.Mol(mol)
			
		self._mol = mol

		self._db = db
		
	### FACTORIES

	### PROPERTIES

	@property
	def id(self) -> int:
		"""Returns the compound's database ID"""
		return self._id
	
	@property
	def inchikey(self) -> str:
		"""Returns the compound's InChiKey"""
		return self._inchikey
	
	@property
	def name(self) -> str:
		"""Returns the compound's InChiKey"""
		return self._inchikey
	
	@property
	def smiles(self) -> str:
		"""Returns the compound's (flattened) smiles"""
		return self._smiles

	@property
	def alias(self) -> str:
		"""Returns the compound's alias"""
		return self._alias

	@alias.setter
	def alias(self, alias: str) -> None:
		"""Set the compound's alias"""
		assert isinstance(alias, str)
		self._alias = alias
		self.db.update(table='compound', id=self.id, key='compound_alias', value=alias, commit=commit)
	
	@property
	def mol(self) -> Chem.Mol:
		"""Returns the compound's RDKit Molecule"""
		if self._mol is None:
			mol, = self.db.select_where(query='mol_to_binary_mol(compound_mol)', table='compound', key='id', value=self.id, multiple=False)
			self._mol = Chem.Mol(mol)
		return self._mol

	@property
	def num_heavy_atoms(self):
		"""Get the number of heavy atoms"""
		if self._num_heavy_atoms is None:
			self._num_heavy_atoms = self.db.get_compound_computed_property('num_heavy_atoms', self.id)
		return self._num_heavy_atoms

	@property
	def molecular_weight(self):
		"""Get the molecular weight"""
		if self._molecular_weight is None:
			self._molecular_weight = self.db.get_compound_computed_property('molecular_weight', self.id)
		return self._molecular_weight

	@property
	def num_rings(self):
		"""Get the number of rings"""
		if self._num_rings is None:
			self._num_rings = self.db.get_compound_computed_property('num_rings', self.id)
		return self._num_rings

	@property
	def formula(self):
		"""Get the chemical formula"""
		if self._formula is None:
			self._formula = self.db.get_compound_computed_property('formula', self.id)
		return self._formula

	@property
	def atomtype_dict(self):
		from .tools import formula_to_atomtype_dict
		return formula_to_atomtype_dict(self.formula)

	@property
	def num_atoms_added(self):
		"""Calculate the number of atoms added relative to the base compound"""
		assert (b_id := self._base_id), f'{self} has no base defined'
		n_e = self.num_heavy_atoms
		n_b = self.db.get_compound_computed_property('num_heavy_atoms', b_id)
		return n_e - n_b
	
	@property
	def metadata(self) -> dict:
		"""Returns the compound's metadata dict"""
		if self._metadata is None:
			self._metadata = self.db.get_metadata(table='compound', id=self.id)
		return self._metadata

	@property
	def db(self):
		"""Returns a pointer to the parent database"""
		return self._db

	@property
	def tags(self) -> TagSet:
		"""Returns the compound's tags"""
		if not self._tags:
			self._tags = self.get_tags()
		return self._tags

	@property
	def poses(self):
		"""Returns the compound's poses"""
		return self.get_poses()

	@property
	def best_placed_pose(self):
		return self.poses.best_placed_pose

	@property
	def num_poses(self) -> int:
		"""Returns the number of associated poses"""
		return self.db.count_where(table='pose', key='compound', value=self.id)

	@property
	def num_reactions(self) -> int:
		"""Returns the number of associated reactions (product)"""
		return self.db.count_where(table='reaction', key='product', value=self.id)

	@property
	def num_reactant(self) -> int:
		"""Returns the number of associated reactions (reactant)"""
		return self.db.count_where(table='reactant', key='compound', value=self.id)

	@property
	def base(self):
		"""Returns the base compound for this elaboration"""
		if self._base_id and self._base is None:
			self._base = self.db.get_compound(id=self._base_id)
		return self._base

	@base.setter
	def base(self, b):
		"""Set the base compound for this elaboration"""
		self.set_base(b)

	@property
	def reactions(self):
		"""Returns the reactions resulting in this compound"""
		return self.get_reactions(none=False)

	@property
	def dict(self) -> dict:
		"""Returns a dictionary of this compound"""
		return self.get_dict()

	@property
	def table(self):
		return self._table

	@property
	def elabs(self):
		"""Return the set of elaborations based on this compound"""
		ids = self.db.select_where(query='compound_id', table='compound', key='base', value=self.id, multiple=True)
		ids = [q for q, in ids]
		from .cset import CompoundSet
		return CompoundSet(self.db, ids)

	@property
	def is_base(self):
		"""Is this Compound the basis for any elaborations?"""
		return bool(self.db.select_where(query='compound_id', table='compound', key='base', value=self.id, multiple=False, none='quiet'))

	@property
	def is_elab(self):
		"""Is this Compound the based on any other compound?"""
		return bool(self.base)

	@property
	def is_product(self):
		"""Is this Compound a product of at least one reaction"""
		return bool(self.get_reactions(none=False))
	
	
	### METHODS

	# def get_dict(self, mol=False):
	# 	"""Returns a dictionary of this compound"""

	# 	serialisable_fields = ['id','inchikey','alias','smiles']

	# 	if mol:
	# 		serialisable_fields.append('mol')

	# 	data = {}
	# 	for key in serialisable_fields:
	# 		data[key] = getattr(self, key)

	# 	if base := self.base:
	# 		data['base'] = base.inchikey

	# 	if metadata := self.metadata:
	# 		for key in metadata:
	# 			data[key] = metadata[key]

	# 	return data

	def add_stock(self, 
		amount: float,
		*,
		purity: float | None = None,
		entry: str | None = None,
		location:str | None = None,
		return_quote = True,
	) -> int | Quote:

		"""Register a certain quantity of compound stock in the Database."""

		assert amount

		# search for existing in stock quotes
		existing = self.get_quotes(supplier='Stock', df=False)

		# supersede old in stock records
		if existing:
			delete = set()
			not_deleted = 0
			for quote in existing:

				if any([
					quote.entry != entry,
					quote.purity != purity,
					quote.catalogue != location,
				]):
					not_deleted += 1
					continue

				delete.add(quote.id)

			delete_str = str(tuple(delete)).replace(',)',')')

			self.db.delete_where(table='quote', key=f'quote_id IN {delete_str}')

			if delete:
				logger.warning(f'Removed {len(delete)} existing In-Stock Quotes')
			
			if not_deleted:
				logger.warning(f'Did not remove {not_deleted} existing In-Stock Quotes with differing entry/purity/location')

		# insert the new quote
		quote_id = self.db.insert_quote(
			compound=self.id,
			price=0,
			lead_time=0,
			currency=None,
			supplier='Stock',
			catalogue=location,
			entry=entry,
			amount=amount,
			purity=purity,
		)

		if return_quote:
			self.db.get_quote(id=quote_id)
		else:
			return quote_id

	def get_tags(self) -> set:
		tags = self.db.select_where(query='tag_name', table='tag', key='compound', value=self.id, multiple=True, none='quiet')
		return TagSet(self, {t[0] for t in tags}, commit=False)

	def get_quotes(self, min_amount=None, supplier=None, max_lead_time=None, none='quiet', pick_cheapest=False, df=False) -> list[dict]:
		"""Get all quotes associated to this compound"""

		quote_ids = self.db.select_where(query='quote_id', table='quote', key='compound', value=self.id, multiple=True, none=none)

		if quote_ids:
			quotes = [self.db.get_quote(id=q[0]) for q in quote_ids]
		else:
			return None

		if supplier:
			quotes = [q for q in quotes if q.supplier == supplier]

		if max_lead_time:
			quotes = [q for q in quotes if q.lead_time <= max_lead_time]

		if min_amount:
			suitable_quotes = [q for q in quotes if q.amount >= min_amount]
			
			if not suitable_quotes:
				# logger.debug(f'No quote available with amount >= {min_amount} mg')
				quotes = [Quote.combination(min_amount, quotes)]

			else:
				quotes = suitable_quotes

		if pick_cheapest:
			return sorted(quotes, key=lambda x: x.price)[0]

		if df:
			from pandas import DataFrame
			return DataFrame([q.dict for q in quotes]).drop(columns='compound')
		
		return quotes

	def get_reactions(self, none='error', as_reactant=False, permitted_reactions=None) -> list:
		"""Get the associated reactions as product, unless as_reactant is True."""

		from .rset import ReactionSet

		if as_reactant:
			reaction_ids = self.db.select_where(query='reactant_reaction', table='reactant', key='compound', value=self.id, multiple=True, none=none)
		else:
			reaction_ids = self.db.select_where(query='reaction_id', table='reaction', key='product', value=self.id, multiple=True, none=none)

		reaction_ids = [q for q, in reaction_ids]

		if permitted_reactions:
			reaction_ids = [i for i in reaction_ids if i in permitted_reactions]

		return ReactionSet(self.db, reaction_ids)

	def get_poses(self, 
		target: str = None
	) -> list[Pose]:

		pose_ids = self.db.select_where(query='pose_id', table='pose', key='compound', value=self.id, multiple=True, none=False)

		# if not pose_ids:
			# return None

		# poses = [self.db.get_pose(id=q[0]) for q in pose_ids]

		from .pset import PoseSet

		return PoseSet(self.db, [q[0] for q in pose_ids])

	def get_dict(self, mol=True, reactions=False, metadata=True, count_by_target=False, poses=True):
		
		"""Returns a dictionary representing this Compound"""

		# serialisable_fields = ['id','alias', 'inchikey', 'smiles', 'num_poses', 'num_reactant', 'num_reactions']
		serialisable_fields = ['id','alias', 'inchikey', 'smiles', 'num_reactant', 'num_reactions']

		# poses
		# reactions
		# poses.targets

		# assert not reactions

		data = {}
		for key in serialisable_fields:
			data[key] = getattr(self, key)

		if mol:
			try:
				data['mol'] = self.mol
			except InvalidMolError:
				data['mol'] = None

		if self.base:
			data['base'] = self.base.name
		else:
			data['base'] = None

		data['tags'] = self.tags
		
		if poses:

			poses = self.poses
			
			if poses:

				# data['poses'] = [a if a else i for a,i in zip(poses.aliases, poses.inchikeys)]
				# data['poses'] = self.poses.names
				data['poses'] = poses.ids
				data['targets'] = poses.target_names
			
				if count_by_target:
					target_ids = poses.target_ids

					for target in self._animal.targets:
						t_poses = poses(target=target.id) or []
						data[f'#poses {target.name}'] = len(t_poses)
		
		if metadata and (metadict := self.metadata):
			for key in metadict:
				data[key] = metadict[key]

		return data

	def set_base(self, base, commit=True):
		if not isinstance(base, int):
			assert base._table == 'compound'
			base = base.id
		self._base_id = base
		self.db.update(table='compound', id=self.id, key='compound_base', value=base, commit=commit)

	def as_ingredient(self, amount, max_lead_time=None, supplier=None):
		"""Convert this compound into an Ingredient object"""
		
		quote = self.get_quotes(
			pick_cheapest=True, 
			min_amount=amount, 
			max_lead_time=max_lead_time, 
			supplier=supplier,
			none='quiet',
		)

		if not quote:
			quote = None

		# else:
			# quote = quote.id
		
		return Ingredient(self.db, self.id, amount, quote, supplier, max_lead_time)

	def draw(self, align_substructure: bool = False):
		"""Display this compound (and its base if it has one)"""

		if (base := self.base):

			# print(self.base)

			from molparse.rdkit import draw_mcs

			data = {self.base.smiles:f'{base} (base)', self.smiles:str(self)}

			if len(data) == 2:

				# print(data)

				drawing = draw_mcs(data, align_substructure=align_substructure, show_mcs=False, highlight_common=False)
				display(drawing)

			else:
				logger.error(f'Problem drawing {base.id=} vs {self.id=}, self referential?')
				display(self.mol)
		
		else:
			display(self.mol)

	def summary(self, metadata: bool = True, draw: bool = True):
		"""Print a summary of this compound"""
		logger.header(repr(self))

		logger.var('inchikey', self.inchikey)
		logger.var('alias', self.alias)
		logger.var('smiles', self.smiles)
		logger.var('base', self.base)
		
		logger.var('is_base', self.is_base)
		logger.var('num_heavy_atoms', self.num_heavy_atoms)
		logger.var('num_rings', self.num_rings)
		logger.var('formula', self.formula)

		poses = self.poses
		logger.var('#poses', len(poses))
		if poses:
			logger.var('targets', poses.targets)
		
		logger.var('#reactions (product)', self.num_reactions)
		logger.var('#reactions (reactant)', self.num_reactant)
		
		logger.var('tags', self.tags)
		
		if metadata:
			logger.var('metadata', str(self.metadata))
		
		if draw:
			return self.draw()

	def place(self,
		*,
		target: str | int | Target | None = None,
		inspirations: list[Pose] | None = None,
		reference: Pose | None = None,
		max_ddG: float = 0.0,
		max_RMSD: float = 2.0,
		output_dir: str = 'wictor_place',
		tags = None,
		metadata = None,
		overwrite = False,
	) -> Pose:
		"""Generate a new pose for this compound using Fragmenstein."""
		
		from fragmenstein import Monster, Wictor
		from pathlib import Path

		tags = tags or []
		metadata = metadata or {}

		# get required data
		smiles = self.smiles

		inspirations = inspirations or self.poses[0].inspirations
		target = target or self.poses[0].target.name
		reference = reference or self.poses[0].reference

		inspiration_mols = [c.mol for c in inspirations]
		protein_pdb_block = reference.protein_system.pdb_block_with_alt_sites
				
		# create the victor
		victor = Wictor(hits=inspiration_mols, pdb_block=protein_pdb_block)
		victor.work_path = output_dir
		victor.enable_stdout(logging.CRITICAL)

		# do the placement
		victor.place(smiles, long_name=self.name)

		# metadata
		metadata['ddG'] = victor.energy_score['bound']['total_score'] - victor.energy_score['unbound']['total_score']
		metadata['RMSD'] = victor.mrmsd.mrmsd

		# print(victor.energy_score)

		if metadata['ddG'] > max_ddG:
			return None

		if metadata['RMSD'] > max_RMSD:
			return None

		# register the pose
		pose = self._animal.register_pose(
			compound=self,
			target=target,
			path=Path(victor.work_path) / self.name / f'{self.name}.minimised.mol',
			inspirations = inspirations,
			reference=reference,
			tags=tags,
			metadata=metadata,	
		)

		if overwrite:
			ids = [p.id for p in self.poses if p.id != pose.id]
			for i in ids:
				self.db.delete_where(table='pose', key="id", value=i)
			logger.success(f'Successfully posed {self} (and deleted old poses)')
		else:
			logger.success(f'Successfully posed {self}')

		return pose

	### DUNDERS

	def __str__(self):
		return f'C{self.id}'

	def __repr__(self):
		return f'{mcol.bold}{mcol.underline}{self} "{self.name}"{mcol.unbold}{mcol.ununderline}'

	def __eq__(self, other):
		return self.id == other.id

class Ingredient:

	"""An ingredient is a :class:`.Compound` with a fixed quanitity and an attached quote.

	Create one from a :meth:`.Compound.as_ingredient`"""

	_table = 'ingredient'

	def __init__(self, db, compound, amount, quote, max_lead_time=None, supplier=None):

		assert compound

		self._db = db
			
		# don't store inherited compound in memory until needed
		self._compound = None
	
		if isinstance(compound, Compound):
			self._compound_id = compound.id
			self._compound = None
		else:
			self._compound_id = compound

		if isinstance(quote, Quote):

			if id := quote.id:
				self._quote_id = quote.id
				self._quote = None

			else:
				# logger.debug('Initialising Ingredient with estimated quote')
				self._quote_id = None
				self._quote = quote

		elif quote is None:
			self._quote_id = None
			self._quote = None
		
		else:
			self._quote_id = int(quote)
			self._quote = None
		
		# self._id = inherit.id
		# self._inchikey = inherit.inchikey
		# self._alias = inherit.alias
		# self._smiles = inherit.smiles
		# self._base = inherit.base			
		# self._mol = inherit.mol
		# self._db = inherit.db
		
		self._amount = amount
		self._max_lead_time = max_lead_time
		self._supplier = supplier

	### PROPERTIES

	@property
	def db(self):
		return self._db
	
	@property
	def amount(self) -> float:
		"""Returns the amount"""
		return self._amount

	@property
	def id(self):
		return self._compound_id

	@property
	def compound_id(self):
		"""Returns the ID of the associated compound"""
		return self._compound_id

	@property
	def quote_id(self):
		"""Returns the ID of the associated quote"""
		return self._quote_id

	@property
	def max_lead_time(self) -> float:
		"""Returns the max_lead_time from the original quote query"""
		return self._max_lead_time

	@property
	def supplier(self) -> float:
		"""Returns the supplier from the original quote query"""
		return self._supplier

	@amount.setter
	def amount(self, a):
		"""Set the amount and update quotes"""

		quote_id = self.get_cheapest_quote_id( 
			min_amount=a, 
			max_lead_time=self._max_lead_time, 
			supplier=self._supplier,
			none='quiet',
		)

		self._quote_id = quote_id 
		
		self._amount = a

	@property
	def compound(self):
		"""Returns the associated :class:`Compound`"""
		if not self._compound:
			self._compound = self.db.get_compound(id=self.compound_id)
		return self._compound
		
	@property
	def quote(self):
		"""Returns the associated :class:`Quote`"""
		if not self._quote:
			if (q_id := self.quote_id):
				self._quote = self.db.get_quote(id=self.quote_id)
			
			else:
				self._quote = self.compound.get_quotes(
					pick_cheapest=True, 
					min_amount=self.amount, 
					max_lead_time=self.max_lead_time, 
					supplier=self.supplier,
					none='quiet',
				)

		return self._quote

	@property
	def compound_price_amount_str(self):
		# return f'{self} {self.quote.currency_symbol}{self.quote.price} ({self.amount})'
		return f'{self} ({self.amount})'

	@property
	def smiles(self):
		return self.compound.smiles

	@property
	def price(self):
		if self.quote:
			return self.quote.price
		else:
			return None

	@property
	def lead_time(self):
		if self.quote:
			return self.quote.lead_time
		else:
			return None

	### METHODS

	def get_cheapest_quote_id(self, min_amount=None, supplier=None, max_lead_time=None, none='quiet') -> list[dict]:
		"""Query quotes associated to this ingredient"""

		supplier_str = f' AND quote_supplier IS "{supplier}"' if supplier else ""
		lead_time_str = f' AND quote_lead_time <= {max_lead_time}' if max_lead_time else ""
		key_str = f'quote_compound IS {self.compound_id} AND quote_amount >= {min_amount}{supplier_str}{lead_time_str} ORDER BY quote_price'

		result = self.db.select_where(query='quote_id', table='quote', key=key_str, multiple=False, none=none)

		if result:
			quote_id, = result
			return quote_id
		
		else:
			return None

	### DUNDERS

	def __str__(self):
		return f'{self.amount:.2f}mg of C{self._compound_id}'
	
	def __repr__(self):
		return f'{mcol.bold}{mcol.underline}{str(self)}{mcol.unbold}{mcol.ununderline}'

	def __eq__(self, other):

		if self.compound_id != other.compound_id:
			return False

		return self.amount == other.amount




import re
import numpy as np
from molparse.rdkit import mol_from_smiles
from rdkit.Chem.inchi import MolToInchiKey
from rdkit.Chem import MolFromSmiles, MolToSmiles, AddHs, RemoveHs
import mcol
import mout

from mlog import setup_logger
logger = setup_logger('HIPPO')

from ase.data import chemical_symbols
chemical_symbols = chemical_symbols[1:]
chemical_symbols = sorted(chemical_symbols, key=lambda x: len(x), reverse=True)

def df_row_to_dict(df_row):

	assert len(df_row) == 1, f'{len(df_row)=}'

	data = {}

	for col in df_row.columns:

		if col == 'Unnamed: 0':
			continue

		value = df_row[col].values[0]

		if not isinstance(value,str) and np.isnan(value):
			value = None

		data[col] = value

	return data

def remove_other_ligands(sys, residue_number, chain):
	
	ligand_residues = [r.number for r in sys['rLIG'] if r.number != residue_number]
	
	# if ligand_residues:
	for c in sys.chains:
		if c.name != chain:
			c.remove_residues(names=['LIG'], verbosity=0)
		elif ligand_residues:
			c.remove_residues(numbers=ligand_residues, verbosity=0)

	# print([r.name_number_str for r in sys['rLIG']])

	assert len([r.name_number_str for r in sys['rLIG']]) == 1, f"{sys.name} {[r.name_number_str for r in sys['rLIG']]}"

	return sys

def inchikey_from_smiles(smiles):
	mol = mol_from_smiles(smiles)
	return MolToInchiKey(mol)

def flat_inchikey(smiles):
	smiles = sanitise_smiles(smiles)
	return inchikey_from_smiles(smiles)

def remove_isotopes_from_smiles(smiles):

	mol = MolFromSmiles(smiles)

	atom_data = [(atom, atom.GetIsotope()) for atom in mol.GetAtoms()]

	for atom, isotope in atom_data:
		if isotope:
			atom.SetIsotope(0)

	return MolToSmiles(mol)

def smiles_has_isotope(smiles, regex=True):
	if regex:
		return re.search(r'([\[][0-9]+[A-Z]+\])',smiles)
	else:
		mol = MolFromSmiles(smiles)
		return any(atom.GetIsotope() for atom in mol.GetAtoms())

def sanitise_smiles(s, verbosity=False, sanitisation_failed='error', radical='error'):

	assert isinstance(s, str), f'non-string smiles={s}'

	orig_smiles = s

	# if multiple molecules take the largest
	if '.' in s:
		s = sorted(s.split('.'), key=lambda x: len(x))[-1]
	
	# flatten the smiles
	stereo_smiles = s
	smiles = s.replace('@','')
	smiles = smiles.replace('/','')
	smiles = smiles.replace('\\','')

	# remove isotopic stuff
	if smiles_has_isotope(smiles):
		mout.warning(f'Isotope(s) in SMILES: {smiles}')
		smiles = remove_isotopes_from_smiles(smiles)

	# canonicalise
	mol = MolFromSmiles(smiles)
	if mol:
		smiles = MolToSmiles(mol,True)
	elif sanitisation_failed =='error':
		raise SanitisationError
	elif sanitisation_failed =='warning':
		logger.warning(f'sanitisation failed for {smiles=}')

	# check radicals
	reconstruct = False
	for atom in mol.GetAtoms():
		if not atom.GetNumRadicalElectrons():
			continue

		if radical == 'warning':
			logger.warning(f'Radical atom in {smiles=}')
		elif radical == 'error':
			raise SanitisationError(f'Radical atom in {smiles=}')
		elif radical == 'remove':
			logger.warning(f'Removed radical atom')
			atom.SetNumRadicalElectrons(0)         
			smiles = MolToSmiles(mol,True)
			reconstruct = True
			# atom.SetFormalCharge(0)
		else:
			raise NotImplementedError(f'Unknown option {radical=}')

	if reconstruct:
		mol = AddHs(mol)
		mol = RemoveHs(mol, implicitOnly=True)
		smiles = MolToSmiles(mol,True)
		logger.warning(f'New {smiles=}')

	if verbosity:

		if smiles != orig_smiles:

			annotated_smiles_str = orig_smiles.replace('.',f'{mcol.error}{mcol.underline}.{mcol.clear}{mcol.warning}')
			annotated_smiles_str = annotated_smiles_str.replace('@',f'{mcol.error}{mcol.underline}@{mcol.clear}{mcol.warning}')
			
			mout.warning(f'SMILES was changed: {annotated_smiles_str} --> {smiles}')

	return smiles

def sanitise_mol(m):
	from rdkit.Chem import MolToMolBlock, MolFromMolBlock
	return MolFromMolBlock(MolToMolBlock(m))

def pose_gap(a, b):

	from numpy.linalg import norm
	from molparse.rdkit import mol_to_AtomGroup

	min_dist = None

	a = mol_to_AtomGroup(a.mol)
	b = mol_to_AtomGroup(b.mol)

	for atom1 in a.atoms:
		for atom2 in b.atoms:
			dist = norm(atom1.np_pos - atom2.np_pos)
			if min_dist is None or dist < min_dist:
				min_dist = dist

	return min_dist

def formula_to_atomtype_dict(formula):

	import re

	remaining = formula.replace('+','')

	atomtype_dict = {}
	
	while remaining:
		for symbol in chemical_symbols:
			if remaining.startswith(symbol):
				remaining = remaining.removeprefix(symbol)

				count_str = re.findall(r'^([0-9]*)', remaining)[0]

				if count_str:
					remaining = remaining.removeprefix(count_str)
					count = int(count_str)
				else:
					count = 1

				atomtype_dict[symbol] = count

				break
		else:
			raise Exception(f'Could not match any chemical symbol to the start of {remaining} {formula=}')

	return atomtype_dict

def combine_atomtype_dicts(atomtype_dicts):

	merged = {}
	for atomtype_dict in atomtype_dicts:
		for symbol, count in atomtype_dict.items():
			if symbol not in merged:
				merged[symbol] = 0

			merged[symbol] += count

	return merged

def atomtype_dict_to_formula(atomtype_dict):
	
	symbols = []

	"""For organic compounds, the order is carbon, hydrogen, 
	then all other elements in alphabetical order of their chemical symbols."""

	key = 'C'
	if key in atomtype_dict:
		value = atomtype_dict[key]
		if value > 1:
			symbols.append(f'C{value}')
		else:
			symbols.append('C')

	key = 'H'
	if key in atomtype_dict:
		value = atomtype_dict[key]
		if value > 1:
			symbols.append(f'{key}{value}')
		else:
			symbols.append(key)

	keys = sorted(list(atomtype_dict.keys()))

	for key in keys:
		value = atomtype_dict[key]
		if key in ['C', 'H']:
			continue
		if value > 1:
			symbols.append(f'{key}{value}')
		else:
			symbols.append(key)

	return ''.join(symbols)

class SanitisationError(Exception):
	...
